class AgentErrors(Exception):
    pass


class AgentDynamicTraitsFunctionError(AgentErrors):
    pass


class AgentDirectAnswerFunctionError(AgentErrors):
    pass


class AgentAttributeLookupCallbackError(AgentErrors):
    pass


class AgentCombinationError(AgentErrors):
    pass


class AgentLacksLLMError(AgentErrors):
    pass


class AgentRespondedWithBadJSONError(AgentErrors):
    pass


class AgentNameError(AgentErrors):
    pass


class AgentTraitKeyError(AgentErrors):
    pass


class FailedTaskException(Exception):
    def __init__(self, message, agent_response_dict):
        super().__init__(message)
        self.agent_response_dict = agent_response_dict



import functools
import asyncio
import nest_asyncio

nest_asyncio.apply()


def jupyter_nb_handler(func):
    """Decorator to run an async function in the event loop if it's running, or synchronously otherwise."""

    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        # This is an async wrapper to await the coroutine
        return await func(*args, **kwargs)

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If the loop is running, schedule the coroutine and wait for the result
            future = asyncio.ensure_future(async_wrapper(*args, **kwargs))
            return loop.run_until_complete(future)
        else:
            # If the loop is not running, run the coroutine to completion
            return asyncio.run(async_wrapper(*args, **kwargs))

    return wrapper


def sync_wrapper(async_func):
    """Decorator to create a synchronous wrapper for an asynchronous function."""

    @functools.wraps(async_func)
    def wrapper(*args, **kwargs):
        return asyncio.run(async_func(*args, **kwargs))

    return wrapper



"""This module provides a factory class for creating question objects."""
import textwrap
from typing import Union

from edsl.exceptions import QuestionSerializationError
from edsl.exceptions import QuestionCreationValidationError
from edsl.questions.QuestionBase import RegisterQuestionsMeta


class Meta(type):
    """Metaclass for QuestionBase that provides a __repr__ method that lists all available questions."""

    def __repr__(cls):
        """Return a string that lists all available questions."""

        s = textwrap.dedent(
            """
        You can use the Question class to create objects by name. 
        For example, to create a multiple choice question, you can do:

        >>> from edsl import Question
        >>> q = Question('multiple_choice', question_text='What is your favorite color?', question_name='color')
        
        Question Types:\n"""
        )
        for question_type, question_class in cls.available(
            show_class_names=True
        ).items():
            line_info = (
                f"{question_type} ({question_class.__name__}): {question_class.__doc__}"
            )
            s += line_info + "\n"
        return s


class Question(metaclass=Meta):
    """Factory class for creating question objects."""

    def __new__(cls, question_type, *args, **kwargs):
        """Create a new question object."""
        get_question_classes = RegisterQuestionsMeta.question_types_to_classes()

        subclass = get_question_classes.get(question_type, None)
        if subclass is None:
            raise ValueError(
                f"No question registered with question_type {question_type}"
            )

        # Create an instance of the selected subclass
        instance = object.__new__(subclass)
        instance.__init__(*args, **kwargs)
        return instance

    @classmethod
    def pull(cls, id_or_url: str):
        """Pull the object from coop."""
        from edsl.coop import Coop

        c = Coop()
        if c.url in id_or_url:
            id = id_or_url.split("/")[-1]
        else:
            id = id_or_url
        from edsl.questions.QuestionBase import QuestionBase

        return c._get_base(QuestionBase, id)

    @classmethod
    def available(cls, show_class_names: bool = False) -> Union[list, dict]:
        """Return a list of available question types.

        :param show_class_names: If True, return a dictionary of question types to class names. If False, return a set of question types.

        Example usage:

        >>> from edsl import Question
        >>> Question.available()
        ['budget', 'checkbox', 'extract', 'free_text', 'functional', 'likert_five', 'linear_scale', 'list', 'multiple_choice', 'numerical', 'rank', 'top_k', 'yes_no']
        """
        if show_class_names:
            return RegisterQuestionsMeta.question_types_to_classes()
        else:
            return sorted(set(RegisterQuestionsMeta.question_types_to_classes().keys()))


def get_question_class(question_type):
    """Return the class for the given question type."""
    q2c = RegisterQuestionsMeta.question_types_to_classes()
    if question_type not in q2c:
        raise ValueError(
            f"The question type, {question_type}, is not recognized. Recognied types are: {q2c.keys()}"
        )
    return q2c.get(question_type)


question_purpose = {
    "multiple_choice": "When options are known and limited",
    "free_text": "When options are unknown or unlimited",
    "checkbox": "When multiple options can be selected",
    "numerical": "When the answer is a single numerical value e.g., a float",
    "linear_scale": "When options are text, but can be ordered, e.g., daily, weekly, monthly, etc.",
    "yes_no": "When the question can be fully answered with either a yes or a no",
    "list": "When the answer should be a list of items",
    "rank": "When the answer should be a ranked list of items",
    "budget": "When the answer should be an amount allocated among a set of options",
    "top_k": "When the answer should be a list of the top k items",
    "likert_five": "When the answer should be a value on the Likert scale from 1 to 5",
    "extract": "When the answer should be information extracted or extrapolated from a text in a given format",
}


if __name__ == "__main__":
    print(Question.available())

    # q = Question("free_text", question_text="How are you doing?", question_name="test")
    # results = q.run()

    q = Question.pull(id=76)



"""Utilities for working with abstract syntax trees (ASTs)."""
import ast


def extract_variable_names(node):
    """Extract variable names from an abstract syntax tree (AST) node."""
    if isinstance(node, ast.Name):
        return [node.id]  # Extract variable name
    elif isinstance(node, ast.BinOp):
        left_names = extract_variable_names(node.left)
        right_names = extract_variable_names(node.right)
        return left_names + right_names
    elif isinstance(node, ast.UnaryOp):
        return extract_variable_names(node.operand)
    elif isinstance(node, ast.Call):
        names = []
        for arg in node.args:
            names.extend(extract_variable_names(arg))
        return names
    else:
        names = []
        for child in ast.iter_child_nodes(node):
            names.extend(extract_variable_names(child))
        return names



# TODO: Move this to a more appropriate location
from edsl.jobs.tokens.TokenPricing import TokenPricing

# , InterviewTokenUsage

pricing = {
    "gpt-3.5-turbo": TokenPricing(
        model_name="gpt-3.5-turbo",
        prompt_token_price_per_k=0.0005,
        completion_token_price_per_k=0.0015,
    ),
    "gpt-4-1106-preview": TokenPricing(
        model_name="gpt-4",
        prompt_token_price_per_k=0.01,
        completion_token_price_per_k=0.03,
    ),
    "test": TokenPricing(
        model_name="test",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "gemini_pro": TokenPricing(
        model_name="gemini_pro",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "llama-2-13b-chat-hf": TokenPricing(
        model_name="llama-2-13b-chat-hf",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "llama-2-70b-chat-hf": TokenPricing(
        model_name="llama-2-70b-chat-hf",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "mixtral-8x7B-instruct-v0.1": TokenPricing(
        model_name="mixtral-8x7B-instruct-v0.1",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
}



"""Module for creating Invigilators, which are objects to administer a question to an Agent."""
import json
from typing import Coroutine, Dict, Any, Optional

from edsl.exceptions import AgentRespondedWithBadJSONError
from edsl.prompts.Prompt import Prompt
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler
from edsl.prompts.registry import get_classes as prompt_lookup
from edsl.data_transfer_models import AgentResponseDict
from edsl.exceptions.agents import FailedTaskException
from edsl.agents.PromptConstructionMixin import PromptConstructorMixin

from edsl.agents.InvigilatorBase import InvigilatorBase


class InvigilatorAI(PromptConstructorMixin, InvigilatorBase):
    """An invigilator that uses an AI model to answer questions."""

    async def async_answer_question(self, failed: bool = False) -> AgentResponseDict:
        """Answer a question using the AI model."""
        params = self.get_prompts() | {"iteration": self.iteration}
        raw_response = await self.async_get_response(**params)
        assert "raw_model_response" in raw_response
        data = {
            "agent": self.agent,
            "question": self.question,
            "scenario": self.scenario,
        }
        raw_response_data = {
            "raw_response": raw_response,
            "raw_model_response": raw_response["raw_model_response"],
        }
        params = data | raw_response_data
        response = self._format_raw_response(**params)
        return AgentResponseDict(**response)

    async def async_get_response(
        self, user_prompt: Prompt, system_prompt: Prompt, iteration: int = 1
    ) -> dict:
        """Call the LLM and gets a response. Used in the `answer_question` method."""
        try:
            response = await self.model.async_get_response(
                user_prompt=user_prompt.text,
                system_prompt=system_prompt.text,
                iteration=iteration,
                cache=self.cache,
            )

        # TODO: I *don't* think we need to delete the cache key here because I think
        # it will not have been set yet; the exception would have been raised before.
        except json.JSONDecodeError as e:
            raise AgentRespondedWithBadJSONError(
                f"Returned bad JSON: {e}"
                f"Prompt: {user_prompt}"
                f"System Prompt: {system_prompt}"
            )

        return response

    def _format_raw_response(
        self, *, agent, question, scenario, raw_response, raw_model_response
    ) -> AgentResponseDict:
        """Return formatted raw response.

        This cleans up the raw response to make it suitable to pass to AgentResponseDict.
        """
        try:
            response = question._validate_answer(raw_response)
        except Exception as e:
            print("Purging the cache key")
            if (
                "raw_model_response" in raw_response
                and "cache_key" in raw_response["raw_model_response"]
            ):
                cache_key = raw_response["raw_model_response"]["cache_key"]
            else:
                cache_key = None
            del self.cache.data[cache_key]
            raise e

        comment = response.get("comment", "")
        answer_code = response["answer"]
        answer = question._translate_answer_code_to_answer(answer_code, scenario)
        raw_model_response = raw_model_response
        data = {
            "answer": answer,
            "comment": comment,
            "question_name": question.question_name,
            "prompts": {k: v.to_dict() for k, v in self.get_prompts().items()},
            "cached_response": raw_response["cached_response"],
            "usage": raw_response.get("usage", {}),
            "raw_model_response": raw_model_response,
        }
        return AgentResponseDict(**data)

    get_response = sync_wrapper(async_get_response)
    answer_question = sync_wrapper(async_answer_question)


class InvigilatorSidecar(InvigilatorAI):
    """An invigilator that presents the 'raw' question to the agent
    & uses a sidecar model to answer questions."""

    async def async_answer_question(self, failed: bool = False) -> AgentResponseDict:
        """Answer a question using the AI model."""
        from edsl import Model

        advanced_model = self.sidecar_model
        simple_model = self.model
        question = self.question
        human_readable_question = (
            "Please answer this single question: " + question.human_readable()
        )
        print("Getting the simple model response to: ", human_readable_question)
        raw_simple_response = await simple_model.async_execute_model_call(
            user_prompt=human_readable_question,
            system_prompt="""Pretend you are a human answering a question. Do not break character.""",
        )
        simple_response = simple_model.parse_response(raw_simple_response)
        instructions = question.get_instructions()

        main_model_prompt = Prompt(
            text="""
        A simpler language model was asked this question: 

        To the simpel model:
        {{ human_readable_question }}

        The simple model responded:
        <response>
        {{ simple_response }}
        </response>

        It was suppose to respond according to these instructions:                                                      
        <instructions>
        {{ instructions }}
        </instructions>
                                
        Please format the simple model's response as it should have been formmated, given the instructions.
        Only respond in valid JSON, like so {"answer": "SPAM!"} or {"answer": "SPAM!", "comment": "I am a robot."}
        Do not inlcude the word 'json'
        """
        )

        d = {
            "human_readable_question": human_readable_question,
            "simple_response": simple_response,
            "instructions": instructions,
        }

        print("The human-readable question is: ", human_readable_question)
        print("The simple response is: ", simple_response)

        raw_response_data = await advanced_model.async_execute_model_call(
            user_prompt=main_model_prompt.render(d).text,
            system_prompt="You are a helpful assistant.",
        )

        raw_response = await advanced_model.async_get_response(
            user_prompt=main_model_prompt.render(d).text,
            system_prompt="You are a helpful assistant.",
            iteration=0,
            cache=self.cache,
        )

        data = {
            "agent": self.agent,
            "question": self.question,
            "scenario": self.scenario,
        }
        raw_response_data = {
            "raw_response": raw_response,
            "raw_model_response": raw_response["raw_model_response"],
        }
        params = data | raw_response_data
        response = self._format_raw_response(**params)
        response.update({"simple_model_raw_response": simple_response})
        return AgentResponseDict(**response)

    # get_response = sync_wrapper(async_get_response)
    answer_question = sync_wrapper(async_answer_question)


class InvigilatorDebug(InvigilatorBase):
    """An invigilator class for debugging purposes."""

    async def async_answer_question(self, iteration: int = 0) -> AgentResponseDict:
        """Return the answer to the question."""
        results = self.question._simulate_answer(human_readable=True)
        results["prompts"] = self.get_prompts()
        results["question_name"] = self.question.question_name
        results["comment"] = "Debug comment"
        return AgentResponseDict(**results)

    def get_prompts(self) -> Dict[str, Prompt]:
        """Return the prompts used."""
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }


class InvigilatorHuman(InvigilatorBase):
    """An invigilator for when a human is answering the question."""

    async def async_answer_question(self, iteration: int = 0) -> AgentResponseDict:
        """Return the answer to the question."""
        data = {
            "comment": "This is a real survey response from a human.",
            "answer": None,
            "prompts": self.get_prompts(),
            "question_name": self.question.question_name,
        }
        try:
            answer = self.agent.answer_question_directly(self.question, self.scenario)
            return AgentResponseDict(**(data | {"answer": answer}))
        except Exception as e:
            agent_response_dict = AgentResponseDict(
                **(data | {"answer": None, "comment": str(e)})
            )
            raise FailedTaskException(
                f"Failed to get response. The exception is {str(e)}",
                agent_response_dict,
            ) from e


class InvigilatorFunctional(InvigilatorBase):
    """A Invigilator for when the question has a answer_question_directly function."""

    async def async_answer_question(self, iteration: int = 0) -> AgentResponseDict:
        """Return the answer to the question."""
        func = self.question.answer_question_directly
        data = {
            "comment": "Functional.",
            "prompts": self.get_prompts(),
            "question_name": self.question.question_name,
        }
        try:
            answer = func(scenario=self.scenario, agent_traits=self.agent.traits)
            return AgentResponseDict(**(data | {"answer": answer}))
        except Exception as e:
            agent_response_dict = AgentResponseDict(
                **(data | {"answer": None, "comment": str(e)})
            )
            raise FailedTaskException(
                f"Failed to get response. The exception is {str(e)}",
                agent_response_dict,
            ) from e

    def get_prompts(self) -> Dict[str, Prompt]:
        """Return the prompts used."""
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }


if __name__ == "__main__":
    pass
#    from edsl.enums import LanguageModelType

# from edsl.agents.Agent import Agent

# a = Agent(
#     instruction="You are a happy-go lucky agent.",
#     traits={"feeling": "happy", "age": "Young at heart"},
#     codebook={"feeling": "Feelings right now", "age": "Age in years"},
#     trait_presentation_template="",
# )

# class MockModel:
#     """Mock model for testing."""

#     model = LanguageModelType.GPT_4.value

# class MockQuestion:
#     """Mock question for testing."""

#     question_type = "free_text"
#     question_text = "How are you feeling?"
#     question_name = "feelings_question"
#     data = {
#         "question_name": "feelings",
#         "question_text": "How are you feeling?",
#         "question_type": "feelings_question",
#     }

# i = InvigilatorAI(
#     agent=a,
#     question=MockQuestion(),
#     scenario={},
#     model=MockModel(),
#     memory_plan=None,
#     current_answers=None,
# )
# print(i.get_prompts()["system_prompt"])
# assert i.get_prompts()["system_prompt"].text == "You are a happy-go lucky agent."

# ###############
# ## Render one
# ###############

# a = Agent(
#     instruction="You are a happy-go lucky agent.",
#     traits={"feeling": "happy", "age": "Young at heart"},
#     codebook={"feeling": "Feelings right now", "age": "Age in years"},
#     trait_presentation_template="You are feeling {{ feeling }}.",
# )

# i = InvigilatorAI(
#     agent=a,
#     question=MockQuestion(),
#     scenario={},
#     model=MockModel(),
#     memory_plan=None,
#     current_answers=None,
# )
# print(i.get_prompts()["system_prompt"])

# assert (
#     i.get_prompts()["system_prompt"].text
#     == "You are a happy-go lucky agent. You are feeling happy."
# )
# try:
#     assert i.get_prompts()["system_prompt"].unused_traits(a.traits) == ["age"]
# except AssertionError:
#     unused_traits = i.get_prompts()["system_prompt"].unused_traits(a.traits)
#     print(f"System prompt: {i.get_prompts()['system_prompt']}")
#     print(f"Agent traits: {a.traits}")
#     print(f"Unused_traits: {unused_traits}")
#     # breakpoint()

# ###############
# ## Render one
# ###############

# a = Agent(
#     instruction="You are a happy-go lucky agent.",
#     traits={"feeling": "happy", "age": "Young at heart"},
#     codebook={"feeling": "Feelings right now", "age": "Age in years"},
#     trait_presentation_template="You are feeling {{ feeling }}. You eat lots of {{ food }}.",
# )

# i = InvigilatorAI(
#     agent=a,
#     question=MockQuestion(),
#     scenario={},
#     model=MockModel(),
#     memory_plan=None,
#     current_answers=None,
# )
# print(i.get_prompts()["system_prompt"])

# ## Should raise a QuestionScenarioRenderError
# assert (
#     i.get_prompts()["system_prompt"].text
#     == "You are a happy-go lucky agent. You are feeling happy."
# )



import textwrap


def get_model_class(model_name, registry=None):
    from edsl.inference_services.registry import default

    registry = registry or default
    factory = registry.create_model_factory(model_name)
    return factory


class Meta(type):
    def __repr__(cls):
        return textwrap.dedent(
            f"""\
        Available models: {cls.available()}
        
        To create an instance, you can do: 
        >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
        
        To get the default model, you can leave out the model name. 
        To see the available models, you can do:
        >>> Model.available()
        """
        )


class Model(metaclass=Meta):
    default_model = "gpt-4-1106-preview"

    def __new__(cls, model_name=None, registry=None, *args, **kwargs):
        # Map index to the respective subclass
        if model_name is None:
            model_name = cls.default_model
        from edsl.inference_services.registry import default

        registry = registry or default
        factory = registry.create_model_factory(model_name)
        return factory(*args, **kwargs)

    @classmethod
    def available(cls, search_term=None, name_only=False, registry=None):
        from edsl.inference_services.registry import default

        registry = registry or default
        full_list = registry.available()
        if search_term is None:
            if name_only:
                return [m[0] for m in full_list]
            else:
                return full_list
        else:
            filtered_results = [
                m for m in full_list if search_term in m[0] or search_term in m[1]
            ]
            if name_only:
                return [m[0] for m in filtered_results]
            else:
                return filtered_results

    @classmethod
    def check_models(cls, verbose=False):
        print("Checking all available models...\n")
        for model in cls.available(name_only=True):
            print(f"Now checking: {model}")
            try:
                m = cls(model)
            except Exception as e:
                print(f"Error creating instance of {model}: {e}")
                continue
            try:
                results = m.hello(verbose)
                if verbose:
                    print(f"Results from model call: {results}")
            except Exception as e:
                print(f"Error calling 'hello' on {model}: {e}")
                continue
            print("OK!")
            print("\n")


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    available = Model.available()
    m = Model("gpt-4-1106-preview")
    results = m.execute_model_call("Hello world")
    print(results)



from __future__ import annotations
import json
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker
from typing import Any, Generator, Optional, Union
from edsl.config import CONFIG
from edsl.data.CacheEntry import CacheEntry
from edsl.data.orm import Base, Data


class SQLiteDict:
    """
    A dictionary-like object that is an interface for an local database.
    - You can use SQLiteDict as a regular dictionary.
    - Supports only SQLite for now.
    """

    def __init__(self, db_path: Optional[str] = None):
        """

        >>> temp_db_path = self._get_temp_path()
        >>> SQLiteDict(f"sqlite:///{temp_db_path}")  # Use the temp file for SQLite
        >>> os.unlink(temp_db_path)  # Clean up the temp file after the test

        """
        self.db_path = db_path or CONFIG.get("EDSL_DATABASE_PATH")
        if not self.db_path.startswith("sqlite:///"):
            self.db_path = f"sqlite:///{self.db_path}"
        try:
            self.engine = create_engine(self.db_path, echo=False, future=True)
            Base.metadata.create_all(self.engine)
            self.Session = sessionmaker(bind=self.engine)
        except SQLAlchemyError as e:
            raise Exception(
                f"""Database initialization error: {e}. The attempted DB path was {db_path}"""
            ) from e

    def _get_temp_path(self):
        import tempfile
        import os

        _, temp_db_path = tempfile.mkstemp(suffix=".db")
        return temp_db_path

    def __setitem__(self, key: str, value: CacheEntry) -> None:
        """
        Stores a key-value pair.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        """
        if not isinstance(value, CacheEntry):
            raise ValueError(f"Value must be a CacheEntry object (got {type(value)}).")
        with self.Session() as db:
            db.merge(Data(key=key, value=json.dumps(value.to_dict())))
            db.commit()

    def __getitem__(self, key: str) -> CacheEntry:
        """
        Gets a value for a given key.
        - Raises a KeyError if the key is not found.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> d["foo"] == CacheEntry.example()
        True
        """
        with self.Session() as db:
            value = db.query(Data).filter_by(key=key).first()
            if not value:
                raise KeyError(f"Key '{key}' not found.")
            return CacheEntry.from_dict(json.loads(value.value))

    def get(self, key: str, default: Optional[Any] = None) -> Union[CacheEntry, Any]:
        """
        Gets the value for a given key
        - Returns the `default` value if the key is not found.

        >>> d = SQLiteDict.example()
        >>> d.get("foo", "bar")
        'bar'
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __bool__(self) -> bool:
        """This is so likes like
        self.data = data or {} 'work' as expected
        """
        return True

    def update(
        self,
        new_d: Union[dict, SQLiteDict],
        overwrite: Optional[bool] = False,
        max_batch_size: Optional[int] = 100,
    ) -> None:
        """
        Update the dictionary with the values from another dictionary.

        :param new_d: The dictionary to update the current dictionary with.
        :param overwrite: If `overwrite` is False, existing values will not be overwritten.
        :param max_batch_size: The maximum number of items to update in a single transaction.

        - If `overwrite` is True, existing values will be overwritten.

        >>> d = SQLiteDict.example()
        >>> d.update({"foo": CacheEntry.example()})
        >>> d["foo"] == CacheEntry.example()
        True
        """
        if not (isinstance(new_d, dict) or isinstance(new_d, SQLiteDict)):
            raise ValueError(
                f"new_d must be a dict or SQLiteDict object (got {type(new_d)})"
            )
        current_batch = 0
        with self.Session() as db:
            for key, value in new_d.items():
                if current_batch == max_batch_size:
                    db.commit()
                    current_batch = 0
                current_batch += 1
                if key in self:
                    if overwrite:
                        db.merge(Data(key=key, value=json.dumps(value.to_dict())))
                else:
                    db.merge(Data(key=key, value=json.dumps(value.to_dict())))
            db.commit()

    def values(self) -> Generator[CacheEntry, None, None]:
        """
        Returns a generator that yields the values in the cache.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> list(d.values()) == [CacheEntry.example()]
        True
        """
        with self.Session() as db:
            for instance in db.query(Data).all():
                yield CacheEntry.from_dict(json.loads(instance.value))

    def items(self) -> Generator[tuple[str, CacheEntry], None, None]:
        """
        Returns a generator that yields the items in the cache.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> list(d.items()) == [("foo", CacheEntry.example())]
        True
        """
        with self.Session() as db:
            for instance in db.query(Data).all():
                yield (instance.key, CacheEntry.from_dict(json.loads(instance.value)))

    def to_dict(self):
        """
        Returns the cache as a dictionary.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> d.to_dict() == {"foo": CacheEntry.example()}
        True
        """
        return dict(self.items())

    def __delitem__(self, key: str) -> None:
        """
        Deletes the value for a given key.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> del d["foo"]
        >>> d.get("foo", "missing")
        'missing'
        """
        with self.Session() as db:
            instance = db.query(Data).filter_by(key=key).one_or_none()
            if instance:
                db.delete(instance)
                db.commit()
            else:
                raise KeyError(f"Key '{key}' not found.")

    def __contains__(self, key: str) -> bool:
        """
        Checks if the dict contains the given key.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> "foo" in d
        True
        >>> "bar" in d
        False
        """
        with self.Session() as db:
            return db.query(Data).filter_by(key=key).first() is not None

    def __iter__(self) -> Generator[str, None, None]:
        """
        Returns a generator that yields the keys in the dict.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> list(iter(d)) == ["foo"]
        True
        """
        with self.Session() as db:
            for instance in db.query(Data).all():
                yield instance.key

    def __len__(self) -> int:
        """
        Returns the number of items in the cache.

        >>> d = SQLiteDict.example()
        >>> len(d)
        0
        >>> d["foo"] = CacheEntry.example()
        >>> len(d)
        1
        """
        with self.Session() as db:
            return db.query(Data).count()

    def keys(self) -> Generator[str, None, None]:
        """
        Returns a generator that yields the keys in the cache.

        >>> d = SQLiteDict.example()
        >>> d["foo"] = CacheEntry.example()
        >>> list(d.keys()) == ["foo"]
        True
        """
        return self.__iter__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(db_path={self.db_path!r})"

    @classmethod
    def example(cls) -> SQLiteDict:
        """
        Returns an example SQLiteDict object.
        - The example SQLiteDict is empty and stored in memory.

        >>> SQLiteDict.example()
        SQLiteDict(db_path='sqlite:///:memory:')
        """
        return cls(db_path="sqlite:///:memory:")


def main():
    from edsl.data.CacheEntry import CacheEntry
    from edsl.data.SQLiteDict import SQLiteDict

    d = SQLiteDict.example()
    d["foo"] = CacheEntry.example()
    d["foo"]
    d.get("foo")
    d.get("poo")
    d.get("poo", "not found")
    d.update({"poo": CacheEntry.example()})
    d["poo"]
    len(d)
    list(d.keys())
    list(d.values())
    list(d.items())
    assert "poo" in d
    assert "loo" not in d
    del d["poo"]
    assert len(d) == 1
    repr(d)
    d


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations
import json
from typing import DefaultDict, Union, List
from collections import defaultdict

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.tasks.task_status_enum import TaskStatus

# for type-hints
from edsl.jobs.runners.JobsRunner import JobsRunner
import asyncio

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]


class InterviewStatusObject:
    pass


class JobsRunHistory:
    """A class to store the history of jobs run.

    Each entry in JobsRunHistory is a tuple:
        - (time, JobStatus). The JobStatus is a list of:
            - InterviewStatus objects (1 per interview) which is a list of
                - Each InterviewStatus object has TaskStatus objects (1 per task)

    It's a list.
    The values is a list of the status updates of each the interviews, with time stamps.

    1. It should be a list of JobStatusObjects, with a time stamp.
    2. A JobStatus object should be a list of InterviewStatusObjects at that time.
    >>> [(time, JobStatusObject), (time, JobStatusObject), (time, JobStatusObject), ...]

    >>> {"interview1": [
        {"time": 0, "status": "InterviewStatusObject"},
        {"time": 1, "status": "InterviewStatusObject"},
        {"time": 2, "status": "InterviewStatusObject"},
        ...
        ],
        "interview2": [
        {"time": 0, "status": "InterviewStatusObject"},
        {"time": 1, "status": "InterviewStatusObject"},
        {"time": 2, "status": "InterviewStatusObject"},
        ...
        ]

    The status at a given time is what entry in the list closest to the asked-for time.

    InterviewStatusObject + InterviewStatusObject --> JobsRunStatusObject

    The goal of this class is to give a detailed history of the jobs run.

    -- Task failures and tracebacks
    -- Response headers from API calls
    -- Log each time an Interview is completed

    Methods:
    --- Visualization tools

    """

    def __init__(self, data: Union[dict, None] = None):
        """Create a new JobsRunHistory object."""
        self.data: dict = data or {}
        self.entries: int = 0
        self.status_functions = {
            "status_dic": self.status_dict,
            "status_counts": self.status_counts,
            "time": self.log_time,
            "exceptions": self.exceptions_dict,
        }

    def print(self):
        """Print the JobsRunHistory object."""
        for elapsed_time, index, status in self.data["status_dic"]:
            # breakpoint()
            print(f"Elapsed Time: {elapsed_time}, Interview index: {index}")
            for interview_status in status:
                interview_status.print()

    def log(
        self,
        JobsRunner: "JobsRunner",
        completed_tasks: List[asyncio.Task],
        elapsed_time: Union[int, float],
    ) -> None:
        """Log the status of the job runner.

        >>> jrh = JobsRunHistory()
        >>> from edsl.jobs.runners.JobsRunnerAsyncio import JobsRunnerAsyncio
        >>> from edsl.jobs.Jobs import Jobs
        >>> jr = JobsRunnerAsyncio(jobs = Jobs.example())
        >>> jrh.log(jr, [], 0)
        >>> jrh.entries
        1
        """

        self.entries += 1

        for name, f in self.status_functions.items():
            entry = f(JobsRunner, completed_tasks, elapsed_time)
            if name not in self.data:
                self.data[name] = []
            self.data[name].append(entry)

    def log_time(self, JobsRunner: "JobsRunner", completed_tasks, elapsed_time):
        return elapsed_time

    def status_dict(
        self, JobsRunner: "JobsRunner", completed_tasks, elapsed_time
    ) -> tuple:
        status = []
        index = -1
        for index, interview in enumerate(JobsRunner.total_interviews):
            status.append(interview.interview_status)
        return (elapsed_time, index, status)

    def exceptions_dict(self, JobsRunner, completed_tasks, elapsed_time) -> tuple:
        exceptions = []
        index = -1
        for index, interview in enumerate(JobsRunner.total_interviews):
            if interview.has_exceptions:
                exceptions.append(interview.exceptions)
        return (elapsed_time, index, exceptions)

    def status_counts(self, JobsRunner, completed_tasks, elapsed_time) -> tuple:
        model_to_status = defaultdict(InterviewStatusDictionary)
        index = -1
        interviews_counted = 0
        for index, interview in enumerate(JobsRunner.total_interviews):
            model = interview.model
            model_to_status[model] += interview.interview_status
            interviews_counted += 1
        assert interviews_counted == len(JobsRunner.total_interviews)
        return (elapsed_time, index, model_to_status)

    def to_dict(self) -> dict:
        d = {}
        for key, value in self.data.items():
            d[key] = [t for t in value]
        return d

    def to_json(self, json_file):
        with open(json_file, "w") as file:
            # json.dump(self.data, file, default=enum_converter)
            json.dump(obj=self.to_dict(), fp=file)

    @classmethod
    def from_json(cls, json_file):
        with open(json_file, "r") as file:
            data = json.load(file)
        return cls(data=data)

    def plot_completion_times(self):
        """Plot the completion times."""
        from matplotlib import pyplot as plt

        # x = [item for item in self.data['time']]

        status_counts = [
            (time, list(d.values())[0]) for time, index, d in self.data["status_counts"]
        ]
        status_counts.sort(key=lambda x: x[0])

        rows = int(len(TaskStatus) ** 0.5) + 1
        cols = (len(TaskStatus) + rows - 1) // rows  # Ensure all plots fit

        plt.figure(figsize=(15, 10))  # Adjust the figure size as needed
        for i, status in enumerate(TaskStatus, start=1):
            plt.subplot(rows, cols, i)
            x = [item[0] for item in status_counts]
            y = [
                item[1].get(status, 0) for item in status_counts
            ]  # Use .get() to handle missing keys safely
            plt.plot(x, y, marker="o", linestyle="-")
            plt.title(status.name)
            plt.xlabel("Elapsed Time")
            plt.ylabel("Count")
            plt.grid(True)

        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    # pass
    import doctest

    doctest.testmod()



from abc import ABC, abstractmethod
import asyncio
from typing import Coroutine, Dict, Any, Optional

from edsl.prompts.Prompt import Prompt
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler
from edsl.prompts.registry import get_classes as prompt_lookup
from edsl.data_transfer_models import AgentResponseDict

from edsl.data.Cache import Cache


class InvigilatorBase(ABC):
    """An invigiator (someone who administers an exam) is a class that is responsible for administering a question to an agent."""

    def __init__(
        self,
        agent,
        question,
        scenario,
        model,
        memory_plan,
        current_answers: dict,
        cache=None,
        iteration: int = 1,
        additional_prompt_data: Optional[dict] = None,
        sidecar_model=None,
    ):
        """Initialize a new Invigilator."""
        self.agent = agent
        self.question = question
        self.scenario = scenario
        self.model = model
        self.memory_plan = memory_plan
        self.current_answers = current_answers
        self.iteration = iteration
        self.additional_prompt_data = additional_prompt_data
        self.cache = cache
        self.sidecar_model = sidecar_model

    def get_failed_task_result(self) -> AgentResponseDict:
        """Return an AgentResponseDict used in case the question-askinf fails."""
        return AgentResponseDict(
            answer=None,
            comment="Failed to get response",
            question_name=self.question.question_name,
            prompts=self.get_prompts(),
        )

    def get_prompts(self) -> Dict[str, Prompt]:
        """Return the prompt used."""
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }

    @classmethod
    def example(cls):
        """Return an example invigilator."""
        from edsl.agents.Agent import Agent
        from edsl.questions import QuestionMultipleChoice
        from edsl.scenarios.Scenario import Scenario
        from edsl.language_models import LanguageModel

        from edsl.enums import InferenceServiceType

        class TestLanguageModelGood(LanguageModel):
            """A test language model."""

            _model_ = "test"
            _parameters_ = {"temperature": 0.5}
            _inference_service_ = InferenceServiceType.TEST.value

            async def async_execute_model_call(
                self, user_prompt: str, system_prompt: str
            ) -> dict[str, Any]:
                await asyncio.sleep(0.1)
                return {"message": """{"answer": "SPAM!"}"""}
                """Return a response from the model."""

            def parse_response(self, raw_response: dict[str, Any]) -> str:
                """Parse the response from the model."""
                return raw_response["message"]

        model = TestLanguageModelGood()
        agent = Agent.example()
        question = QuestionMultipleChoice.example()
        scenario = Scenario.example()
        #        model = LanguageModel.example()
        memory_plan = None
        current_answers = None
        return cls(
            agent=agent,
            question=question,
            scenario=scenario,
            model=model,
            memory_plan=memory_plan,
            current_answers=current_answers,
        )

    @abstractmethod
    async def async_answer_question(self):
        """Asnwer a question."""
        pass

    @jupyter_nb_handler
    def answer_question(self) -> Coroutine:
        """Return a function that gets the answers to the question."""

        async def main():
            """Return the answer to the question."""
            results = await asyncio.gather(self.async_answer_question())
            return results[0]  # Since there's only one task, return its result

        return main()

    def create_memory_prompt(self, question_name):
        """Create a memory for the agent."""
        return self.memory_plan.get_memory_prompt_fragment(
            question_name, self.current_answers
        )



import functools
import json
import requests
from typing import Optional
from edsl.data.CacheEntry import CacheEntry


def handle_request_exceptions(reraise=False):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except requests.exceptions.ConnectionError as e:
                print(f"Could not connect to remote server: {e}")
            except requests.exceptions.Timeout as e:
                print(f"Request timed out: {e}")
            except requests.exceptions.HTTPError as e:
                print(f"HTTP error occurred: {e}")
            except requests.exceptions.RequestException as e:
                print(f"An error occurred during the request: {e}")
            except ValueError as e:
                print(f"Invalid data format: {e}")

            if reraise:
                raise

        return wrapper

    return decorator


class RemoteDict:
    """
    A dictionary-like object that is an interface for a remote database.
    - You can use RemoteDict as a regular dictionary.
    - TODO: Implement the methods.
    """

    def __init__(self, base_url: str = None):
        self.base_url = base_url or "http://127.0.0.1:8000"

    def __bool__(self):
        return True

    def __setitem__(self, key: str, value: CacheEntry):
        response = requests.post(f"{self.base_url}/items/{key}", json=value.to_dict())
        response.raise_for_status()

    def __getitem__(self, key: str):
        response = requests.get(f"{self.base_url}/items/{key}")
        if response.status_code == 404:
            raise KeyError(f"Key '{key}' not found.")
        response.raise_for_status()
        return CacheEntry.from_dict(response.json())

    def get(self, key: str, default: Optional[CacheEntry] = None):
        try:
            return self[key]
        except KeyError:
            return default

    def __delitem__(self, key: str):
        response = requests.delete(f"{self.base_url}/items/{key}")
        if response.status_code == 404:
            raise KeyError(f"Key '{key}' not found.")
        response.raise_for_status()

    def __contains__(self, key: str):
        keys = self.keys()
        return key in keys

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        response = requests.get(f"{self.base_url}/items/count")
        response.raise_for_status()
        return response.json()["count"]

    def keys(self):
        response = requests.get(f"{self.base_url}/items/")
        response.raise_for_status()
        return response.json()

    def values(self) -> list[CacheEntry]:
        response = requests.get(f"{self.base_url}/items/values")
        response.raise_for_status()
        items = response.json()
        return [CacheEntry(**json.loads(item)) for item in items]


if __name__ == "__main__":
    api_dict = RemoteDict()

    # Add an item
    api_dict["example"] = CacheEntry.example()

    # Retrieve an item
    print(api_dict["example"])

    # Check if an item exists
    print("example" in api_dict)



"""Settings for the questions module."""


class Settings:
    """Settings for the questions module."""

    MAX_ANSWER_LENGTH = 2000
    MAX_EXPRESSION_CONSTRAINT_LENGTH = 1000
    MAX_NUM_OPTIONS = 200
    MIN_NUM_OPTIONS = 2
    MAX_OPTION_LENGTH = 1000
    MAX_QUESTION_LENGTH = 100000



"""Enums for the different types of questions, language models, and inference services."""
from enum import Enum


class EnumWithChecks(Enum):
    """Base class for all enums with checks."""

    @classmethod
    def is_value_valid(cls, value):
        """Check if the value is valid."""
        return any(value == item.value for item in cls)


class QuestionType(EnumWithChecks):
    """Enum for the question types."""

    MULTIPLE_CHOICE = "multiple_choice"
    YES_NO = "yes_no"
    FREE_TEXT = "free_text"
    RANK = "rank"
    BUDGET = "budget"
    CHECKBOX = "checkbox"
    EXTRACT = "extract"
    FUNCTIONAL = "functional"
    LIST = "list"
    NUMERICAL = "numerical"
    TOP_K = "top_k"
    LIKERT_FIVE = "likert_five"
    LINEAR_SCALE = "linear_scale"


# https://huggingface.co/meta-llama/Llama-2-70b-chat-hf


# class LanguageModelType(EnumWithChecks):
#     """Enum for the language model types."""

#     GPT_4 = "gpt-4-1106-preview"
#     GPT_3_5_Turbo = "gpt-3.5-turbo"
#     LLAMA_2_70B_CHAT_HF = "llama-2-70b-chat-hf"
#     LLAMA_2_13B_CHAT_HF = "llama-2-13b-chat-hf"
#     GEMINI_PRO = "gemini_pro"
#     MIXTRAL_8x7B_INSTRUCT = "mixtral-8x7B-instruct-v0.1"
#     TEST = "test"
#     ANTHROPIC_3_OPUS = "claude-3-opus-20240229"
#     ANTHROPIC_3_SONNET = "claude-3-sonnet-20240229"
#     ANTHROPIC_3_HAIKU = "claude-3-haiku-20240307"
#     DBRX_INSTRUCT = "dbrx-instruct"


class InferenceServiceType(EnumWithChecks):
    """Enum for the inference service types."""

    BEDROCK = "bedrock"
    DEEP_INFRA = "deep_infra"
    REPLICATE = "replicate"
    OPENAI = "openai"
    GOOGLE = "google"
    TEST = "test"
    ANTHROPIC = "anthropic"


service_to_api_keyname = {
    InferenceServiceType.BEDROCK.value: "TBD",
    InferenceServiceType.DEEP_INFRA.value: "DEEP_INFRA_API_KEY",
    InferenceServiceType.REPLICATE.value: "TBD",
    InferenceServiceType.OPENAI.value: "OPENAI_API_KEY",
    InferenceServiceType.GOOGLE.value: "GOOGLE_API_KEY",
    InferenceServiceType.TEST.value: "TBD",
    InferenceServiceType.ANTHROPIC.value: "ANTHROPIC_API_KEY",
}


class TokenPricing:
    def __init__(
        self,
        *,
        model_name,
        prompt_token_price_per_k: float,
        completion_token_price_per_k: float,
    ):
        self.model_name = model_name
        self.prompt_token_price = prompt_token_price_per_k / 1_000.0
        self.completion_token_price = completion_token_price_per_k / 1_000.0

    def __eq__(self, other):
        if not isinstance(other, TokenPricing):
            return False
        return (
            self.model_name == other.model_name
            and self.prompt_token_price == other.prompt_token_price
            and self.completion_token_price == other.completion_token_price
        )


pricing = {
    "dbrx-instruct": TokenPricing(
        model_name="dbrx-instruct",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "claude-3-opus-20240229": TokenPricing(
        model_name="claude-3-opus-20240229",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "claude-3-haiku-20240307": TokenPricing(
        model_name="claude-3-haiku-20240307",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "claude-3-sonnet-20240229": TokenPricing(
        model_name="claude-3-sonnet-20240229",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "gpt-3.5-turbo": TokenPricing(
        model_name="gpt-3.5-turbo",
        prompt_token_price_per_k=0.0005,
        completion_token_price_per_k=0.0015,
    ),
    "gpt-4-1106-preview": TokenPricing(
        model_name="gpt-4",
        prompt_token_price_per_k=0.01,
        completion_token_price_per_k=0.03,
    ),
    "test": TokenPricing(
        model_name="test",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "gemini_pro": TokenPricing(
        model_name="gemini_pro",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "llama-2-13b-chat-hf": TokenPricing(
        model_name="llama-2-13b-chat-hf",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "llama-2-70b-chat-hf": TokenPricing(
        model_name="llama-2-70b-chat-hf",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
    "mixtral-8x7B-instruct-v0.1": TokenPricing(
        model_name="mixtral-8x7B-instruct-v0.1",
        prompt_token_price_per_k=0.0,
        completion_token_price_per_k=0.0,
    ),
}


def get_token_pricing(model_name):
    if model_name in pricing:
        return pricing[model_name]
    else:
        return TokenPricing(
            model_name=model_name,
            prompt_token_price_per_k=0.0,
            completion_token_price_per_k=0.0,
        )



"""An Agent is an AI agent that can reference a set of traits in answering questions."""
from __future__ import annotations
import copy
import inspect
import types
from typing import Any, Callable, Optional, Union, Dict

from jinja2 import Template
from rich.console import Console
from rich.table import Table

from edsl.Base import Base

from edsl.exceptions.agents import (
    AgentCombinationError,
    AgentDirectAnswerFunctionError,
    AgentDynamicTraitsFunctionError,
)

from edsl.agents.Invigilator import (
    InvigilatorDebug,
    InvigilatorHuman,
    InvigilatorFunctional,
    InvigilatorAI,
    InvigilatorBase,
)

from edsl.language_models.registry import Model
from edsl.scenarios import Scenario

# from edsl.enums import LanguageModelType

from edsl.agents.descriptors import (
    TraitsDescriptor,
    CodebookDescriptor,
    InstructionDescriptor,
    NameDescriptor,
)

from edsl.utilities.decorators import sync_wrapper
from edsl.data_transfer_models import AgentResponseDict
from edsl.prompts.library.agent_persona import AgentPersona

from edsl.data.Cache import Cache


class Agent(Base):
    """An Agent that can answer questions."""

    default_instruction = """You are answering questions as if you were a human. Do not break character."""

    _traits = TraitsDescriptor()
    codebook = CodebookDescriptor()
    instruction = InstructionDescriptor()
    name = NameDescriptor()

    def __init__(
        self,
        *,
        traits: dict = None,
        name: str = None,
        codebook: dict = None,
        instruction: Optional[str] = None,
        traits_presentation_template: Optional[str] = None,
        dynamic_traits_function: Callable = None,
    ):
        """Initialize a new instance of Agent.

        :param traits: A dictionary of traits that the agent has. The keys need to be
        :param name: A name for the agent
        :param codebook: A codebook mapping trait keys to trait descriptions.
        :param instruction: Instructions for the agent in how to answer questions.
        :param trait_presentation_template: A template for how to present the agent's traits.
        :param dynamic_traits_function: A function that returns a dictionary of traits.

        The `traits` parameter is a dictionary of traits that the agent has.
        These traits are used to construct a prompt that is presented to the LLM.
        In the absence of a `traits_presentation_template`, the default is used.
        This is a template that is used to present the agent's traits to the LLM.
        See :py:class:`edsl.prompts.library.agent_persona.AgentPersona` for more information.

        Example usage:

        >>> a = Agent(traits = {"age": 10, "hair": "brown", "height": 5.5})
        >>> a.traits
        {'age': 10, 'hair': 'brown', 'height': 5.5}

        These traits are used to construct a prompt that is presented to the LLM.

        In the absence of a `traits_presentation_template`, the default is used.

        >>> a = Agent(traits = {"age": 10}, traits_presentation_template = "I am a {{age}} year old.")
        >>> repr(a.agent_persona)
        "Prompt(text='I am a {{age}} year old.')"

        When this is rendered for presentation to the LLM, it will replace the `{{age}}` with the actual age.
        it is also possible to use the `codebook` to provide a more human-readable description of the trait.
        Here is an example where we give a prefix to the age trait (namely the age):

        >>> traits = {"age": 10, "hair": "brown", "height": 5.5}
        >>> codebook = {'age': 'Their age is'}
        >>> a = Agent(traits = traits, codebook = codebook, traits_presentation_template = "This agent is Dave. {{codebook['age']}} {{age}}")
        >>> d = a.traits | {'codebook': a.codebook}
        >>> a.agent_persona.render(d)
        Prompt(text='This agent is Dave. Their age is 10')

        Instructions
        ------------------
        The agent can also have instructions. These are instructions that are given to the agent when answering questions.

        >>> Agent.default_instruction
        'You are answering questions as if you were a human. Do not break character.'

        See see how these are used to actually construct the prompt that is presented to the LLM, see :py:class:`edsl.agents.Invigilator.InvigilatorBase`.

        """
        self.name = name
        self._traits = traits or dict()
        self.codebook = codebook or dict()
        self.instruction = instruction or self.default_instruction
        self.dynamic_traits_function = dynamic_traits_function
        self._check_dynamic_traits_function()
        self.current_question = None

        if traits_presentation_template is not None:
            self.traits_presentation_template = traits_presentation_template
            self.agent_persona = AgentPersona(text=self.traits_presentation_template)

    def _check_dynamic_traits_function(self) -> None:
        """Check whether dynamic trait function is valid.

        This checks whether the dynamic traits function is valid.
        """
        if self.dynamic_traits_function:
            sig = inspect.signature(self.dynamic_traits_function)
            if "question" in sig.parameters:
                if len(sig.parameters) > 1:
                    raise AgentDynamicTraitsFunctionError(
                        f"The dynamic traits function {self.dynamic_traits_function} has too many parameters. It should only have one parameter: 'question'."
                    )
            else:
                if len(sig.parameters) > 0:
                    raise AgentDynamicTraitsFunctionError(
                        f"""The dynamic traits function {self.dynamic_traits_function} has too many parameters. It should have no parameters or 
                        just a single parameter: 'question'."""
                    )

    @property
    def traits(self) -> dict[str, str]:
        """An agent's traits, which is a dictionary.

        The agent could have a a dynamic traits function (`dynamic_traits_function`) that returns a dictionary of traits
        when called. This function can also take a `question` as an argument.
        If so, the dynamic traits function is called and the result is returned.
        Otherwise, the traits are returned.

        Example:

        >>> a = Agent(traits = {"age": 10, "hair": "brown", "height": 5.5})
        >>> a.traits
        {'age': 10, 'hair': 'brown', 'height': 5.5}

        """
        if self.dynamic_traits_function:
            sig = inspect.signature(self.dynamic_traits_function)
            if "question" in sig.parameters:
                return self.dynamic_traits_function(question=self.current_question)
            else:
                return self.dynamic_traits_function()
        else:
            return self._traits

    def __getitem__(self, key):
        return getattr(self, key)

    def add_direct_question_answering_method(self, method: Callable) -> None:
        """Add a method to the agent that can answer a particular question type.

        :param method: A method that can answer a question directly.

        Example usage:

        >>> a = Agent()
        >>> def f(self, question, scenario): return "I am a direct answer."
        >>> a.add_direct_question_answering_method(f)
        >>> a.answer_question_directly(question = None, scenario = None)
        'I am a direct answer.'
        """
        if hasattr(self, "answer_question_directly"):
            print("Warning: overwriting existing answer_question_directly method")

        signature = inspect.signature(method)
        for argument in ["question", "scenario", "self"]:
            if argument not in signature.parameters:
                raise AgentDirectAnswerFunctionError(
                    f"The method {method} does not have a '{argument}' parameter."
                )
        bound_method = types.MethodType(method, self)
        setattr(self, "answer_question_directly", bound_method)

    def create_invigilator(
        self,
        *,
        question: Question,
        cache,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
        iteration: int = 1,
        sidecar_model=None,
    ) -> InvigilatorBase:
        """Create an Invigilator.

        An invigator is an object that is responsible for administering a question to an agent and
        recording the responses.
        """
        cache = cache
        self.current_question = question
        model = model or Model()
        scenario = scenario or Scenario()
        invigilator = self._create_invigilator(
            question=question,
            scenario=scenario,
            model=model,
            debug=debug,
            memory_plan=memory_plan,
            current_answers=current_answers,
            iteration=iteration,
            cache=cache,
            sidecar_model=sidecar_model,
        )
        return invigilator

    async def async_answer_question(
        self,
        *,
        question: Question,
        cache,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
        iteration: int = 0,
    ) -> AgentResponseDict:
        """
        Answer a posed question.

        :param question: The question to answer.
        :param scenario: The scenario in which the question is asked.
        :param model: The language model to use.
        :param debug: Whether to run in debug mode.
        :param memory_plan: The memory plan to use.
        :param current_answers: The current answers.
        :param iteration: The iteration number.

        This is a function where an agent returns an answer to a particular question.
        However, there are several different ways an agent can answer a question, so the
        actual functionality is delegated to an Invigilator object.
        """
        invigilator = self.create_invigilator(
            question=question,
            cache=cache,
            scenario=scenario,
            model=model,
            debug=debug,
            memory_plan=memory_plan,
            current_answers=current_answers,
            iteration=iteration,
        )
        response: AgentResponseDict = await invigilator.async_answer_question()
        return response

    answer_question = sync_wrapper(async_answer_question)

    def _create_invigilator(
        self,
        question: Question,
        cache=None,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
        iteration: int = 0,
        sidecar_model=None,
    ) -> InvigilatorBase:
        """Create an Invigilator."""
        model = model or Model()
        scenario = scenario or Scenario()

        if cache is None:
            cache = Cache()

        if debug:
            # use the question's _simulate_answer method
            # breakpoint()
            invigilator_class = InvigilatorDebug
        elif hasattr(question, "answer_question_directly"):
            # it is a functional question and the answer only depends on the agent's traits & the scenario
            invigilator_class = InvigilatorFunctional
        elif hasattr(self, "answer_question_directly"):
            # this of the case where the agent has a method that can answer the question directly
            # this occurrs when 'answer_question_directly' has been monkey-patched onto the agent
            # which happens when the agent is created from an existing survey
            invigilator_class = InvigilatorHuman
        else:
            # this means an LLM agent will be used. This is the standard case.
            invigilator_class = InvigilatorAI

        if sidecar_model is not None:
            from edsl.agents.Invigilator import InvigilatorSidecar

            invigilator_class = InvigilatorSidecar

        invigilator = invigilator_class(
            self,
            question=question,
            scenario=scenario,
            model=model,
            memory_plan=memory_plan,
            current_answers=current_answers,
            iteration=iteration,
            cache=cache,
            sidecar_model=sidecar_model,
        )
        return invigilator

    ################
    # Dunder Methods
    ################
    def __add__(self, other_agent: Agent = None) -> Agent:
        """
        Combine two agents by joining their traits.

        The agents must not have overlapping traits.

        Example usage:

        >>> a1 = Agent(traits = {"age": 10})
        >>> a2 = Agent(traits = {"height": 5.5})
        >>> a1 + a2
        Agent(traits = {'age': 10, 'height': 5.5})
        >>> a1 + a1
        Traceback (most recent call last):
        ...
        edsl.exceptions.agents.AgentCombinationError: The agents have overlapping traits: {'age'}.
        """
        if other_agent is None:
            return self
        elif common_traits := set(self.traits.keys()) & set(other_agent.traits.keys()):
            raise AgentCombinationError(
                f"The agents have overlapping traits: {common_traits}."
            )
        else:
            new_agent = Agent(traits=copy.deepcopy(self.traits))
            new_agent.traits.update(other_agent.traits)
            return new_agent

    def __eq__(self, other: Agent) -> bool:
        """Check if two agents are equal.

        This only checks the traits.
        >>> a1 = Agent(traits = {"age": 10})
        >>> a2 = Agent(traits = {"age": 10})
        >>> a1 == a2
        True
        >>> a3 = Agent(traits = {"age": 11})
        >>> a1 == a3
        False
        """
        return self.data == other.data

    def __repr__(self):
        """Return representation of Agent."""
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()))
        return f"{class_name}({', '.join(items)})"

    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict())

    ################
    # SERIALIZATION METHODS
    ################
    @property
    def data(self):
        """Format the data for serialization.

        TODO: Warn if has dynamic traits function or direct answer function that cannot be serialized.
        TODO: Add ability to have coop-hosted functions that are serializable.
        """
        raw_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        if hasattr(self, "set_instructions"):
            if not self.set_instructions:
                raw_data.pop("instruction")
        if self.codebook == {}:
            raw_data.pop("codebook")
        if self.name == None:
            raw_data.pop("name")
        return raw_data

    def to_dict(self) -> dict[str, Union[dict, bool]]:
        """Serialize to a dictionary."""
        return self.data

    @classmethod
    def from_dict(cls, agent_dict: dict[str, Union[dict, bool]]) -> Agent:
        """Deserialize from a dictionary."""
        return cls(**agent_dict)

    ################
    # DISPLAY Methods
    ################

    def _table(self) -> tuple[dict, list]:
        """Prepare generic table data."""
        table_data = []
        for attr_name, attr_value in self.__dict__.items():
            table_data.append({"Attribute": attr_name, "Value": repr(attr_value)})
        column_names = ["Attribute", "Value"]
        return table_data, column_names

    def rich_print(self):
        """Display an object as a rich table."""
        table_data, column_names = self._table()
        table = Table(title=f"{self.__class__.__name__} Attributes")
        for column in column_names:
            table.add_column(column, style="bold")

        for row in table_data:
            row_data = [row[column] for column in column_names]
            table.add_row(*row_data)

        return table

    @classmethod
    def example(cls) -> Agent:
        """Return an example agent.

        >>> Agent.example()
        Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})
        """
        return cls(traits={"age": 22, "hair": "brown", "height": 5.5})

    def code(self) -> str:
        """Return the code for the agent.
        TODO: Add code for dynamic traits function.
        """
        return f"Agent(traits={self.traits})"


def main():
    """
    Give an example of usage.

    WARNING: Consume API credits
    """
    from edsl.agents import Agent
    from edsl.questions import QuestionMultipleChoice

    # a simple agent
    agent = Agent(traits={"age": 10, "hair": "brown", "height": 5.5})
    agent.traits
    agent.print()
    # combining two agents
    agent = Agent(traits={"age": 10}) + Agent(traits={"height": 5.5})
    agent.traits
    # Agent -> Job using the to() method
    agent = Agent(traits={"allergies": "peanut"})
    question = QuestionMultipleChoice(
        question_text="Would you enjoy a PB&J?",
        question_options=["Yes", "No"],
        question_name="food_preference",
    )
    job = agent.to(question)
    # run the job
    results = job.run()
    # results


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    # a = Agent(
    #     traits={"age": 10}, traits_presentation_template="I am a {{age}} year old."
    # )
    # repr(a.agent_persona)

    # a = Agent(
    #     traits={"age": 22, "hair": "brown", "gender": "female"},
    #     traits_presentation_template="I am a {{ age }} year-old {{ gender }} with {{ hair }} hair.",
    # )
    # print(a.agent_persona.render(primary_replacement=a.traits))



"""This module contains the Question class, which is the base class for all questions in EDSL."""
from __future__ import annotations
from abc import ABC, abstractmethod
from rich.table import Table
from typing import Any, Type, Optional

from edsl.exceptions import (
    QuestionResponseValidationError,
    QuestionSerializationError,
)
from edsl.questions.descriptors import QuestionNameDescriptor, QuestionTextDescriptor

from edsl.prompts.registry import get_classes as prompt_lookup
from edsl.questions.AnswerValidatorMixin import AnswerValidatorMixin
from edsl.questions.RegisterQuestionsMeta import RegisterQuestionsMeta
from edsl.Base import PersistenceMixin, RichPrintingMixin

from edsl.questions.SimpleAskMixin import SimpleAskMixin


class QuestionBase(
    PersistenceMixin,
    RichPrintingMixin,
    SimpleAskMixin,
    ABC,
    AnswerValidatorMixin,
    metaclass=RegisterQuestionsMeta,
):
    """ABC for the Question class. All questions should inherit from this class."""

    question_name: str = QuestionNameDescriptor()
    question_text: str = QuestionTextDescriptor()

    def __getitem__(self, key: str) -> Any:
        """Get an attribute of the question."""
        return getattr(self, key)

    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict())

    @property
    def data(self) -> dict:
        """Return a dictionary of question attributes **except** for question_type."""
        candidate_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        optional_attributes = {
            "set_instructions": "instructions",
        }
        for boolean_flag, attribute in optional_attributes.items():
            if hasattr(self, boolean_flag) and not getattr(self, boolean_flag):
                candidate_data.pop(attribute, None)

        return candidate_data

    @classmethod
    def applicable_prompts(
        cls, model: Optional[str] = None
    ) -> list[type["PromptBase"]]:
        """Get the prompts that are applicable to the question type.

        :param model: The language model to use.

        >>> from edsl.questions import QuestionFreeText
        >>> QuestionFreeText.applicable_prompts()
        [<class 'edsl.prompts.library.question_freetext.FreeText'>]

        :param model: The language model to use. If None, assumes does not matter.

        """
        applicable_prompts = prompt_lookup(
            component_type="question_instructions",
            question_type=cls.question_type,
            model=model,
        )
        return applicable_prompts

    @property
    def model_instructions(self) -> dict:
        """Get the model-specific instructions for the question."""
        if not hasattr(self, "_model_instructions"):
            self._model_instructions = {}
        return self._model_instructions

    @model_instructions.setter
    def model_instructions(self, data: dict):
        """Set the model-specific instructions for the question."""
        self._model_instructions = data

    def add_model_instructions(
        self, *, instructions: str, model: Optional[str] = None
    ) -> None:
        """Add model-specific instructions for the question.

        :param instructions: The instructions to add. This is typically a jinja2 template.
        :param model: The language model for this instruction.

        >>> from edsl.questions import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite color?")
        >>> q.add_model_instructions(instructions = "Answer in valid JSON like so {'answer': 'comment: <>}", model = "gpt3")

        """
        from edsl import Model

        if not hasattr(self, "_model_instructions"):
            self._model_instructions = {}
        if model is None:
            # if not model is passed, all the models are mapped to this instruction, including 'None'
            self._model_instructions = {
                model_name: instructions
                for model_name in Model.available(name_only=True)
            }
            self._model_instructions.update({model: instructions})
        else:
            self._model_instructions.update({model: instructions})

    def get_instructions(self, model: Optional[str] = None) -> type["PromptBase"]:
        """Get the mathcing question-answering instructions for the question.

        :param model: The language model to use.
        """
        from edsl.prompts.Prompt import Prompt

        if model in self.model_instructions:
            return Prompt(text=self.model_instructions[model])
        else:
            return self.applicable_prompts(model)[0]()

    def option_permutations(self) -> list[QuestionBase]:
        """Return a list of questions with all possible permutations of the options."""

        if not hasattr(self, "question_options"):
            return [self]

        import copy
        import itertools

        questions = []
        for index, permutation in enumerate(
            itertools.permutations(self.question_options)
        ):
            question = copy.deepcopy(self)
            question.question_options = list(permutation)
            question.question_name = f"{self.question_name}_{index}"
            questions.append(question)
        return questions

    ############################
    # Serialization methods
    ############################
    def to_dict(self) -> dict[str, Any]:
        """Convert the question to a dictionary that includes the question type (used in deserialization)."""
        candidate_data = self.data.copy()
        candidate_data["question_type"] = self.question_type
        return candidate_data

    @classmethod
    def from_dict(cls, data: dict) -> Type[QuestionBase]:
        """Construct a question object from a dictionary created by that question's `to_dict` method."""
        local_data = data.copy()
        try:
            question_type = local_data.pop("question_type")
            if question_type == "linear_scale":
                # This is a fix for issue https://github.com/expectedparrot/edsl/issues/165
                options_labels = local_data.get("option_labels", None)
                if options_labels:
                    options_labels = {
                        int(key): value for key, value in options_labels.items()
                    }
                    local_data["option_labels"] = options_labels
        except:
            raise QuestionSerializationError(
                f"Data does not have a 'question_type' field (got {data})."
            )
        from edsl.questions.question_registry import get_question_class

        try:
            question_class = get_question_class(question_type)
        except ValueError:
            raise QuestionSerializationError(
                f"No question registered with question_type {question_type}"
            )

        if "model_instructions" in local_data:
            model_instructions = local_data.pop("model_instructions")
            new_q = question_class(**local_data)
            new_q.model_instructions = model_instructions
            return new_q

        return question_class(**local_data)

    ############################
    # Dunder methods
    ############################
    def __repr__(self) -> str:
        """Return a string representation of the question. Should be able to be used to reconstruct the question."""
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()))
        return f"{class_name}({', '.join(items)})"

    def __eq__(self, other: Type[QuestionBase]) -> bool:
        """Check if two questions are equal. Equality is defined as having the .to_dict()."""
        if not isinstance(other, QuestionBase):
            return False
        return self.to_dict() == other.to_dict()

    # TODO: Throws an error that should be addressed at QuestionFunctional
    def __add__(self, other_question):
        """
        Compose two questions into a single question.

        >>> from edsl.scenarios.Scenario import Scenario
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.questions.QuestionNumerical import QuestionNumerical
        >>> q1 = QuestionFreeText(question_text = "What is the capital of {{country}}", question_name = "capital")
        >>> q2 = QuestionNumerical(question_text = "What is the population of {{capital}}, in millions. Please round", question_name = "population")
        >>> q3 = q1 + q2
        """
        from edsl.questions import compose_questions

        return compose_questions(self, other_question)

    @abstractmethod
    def _validate_answer(self, answer: dict[str, str]):
        """Validate the answer from the LLM. Behavior depends on the question type."""
        pass

    def _validate_response(self, response):
        """Validate the response from the LLM. Behavior depends on the question type."""
        if "answer" not in response:
            raise QuestionResponseValidationError(
                "Response from LLM does not have an answer"
            )
        return response

    @abstractmethod
    def _translate_answer_code_to_answer(self):  # pragma: no cover
        """Translate the answer code to the actual answer. Behavior depends on the question type."""
        pass

    @abstractmethod
    def _simulate_answer(self, human_readable=True) -> dict:  # pragma: no cover
        """Simulate a valid answer for debugging purposes (what the validator expects)."""
        pass

    ############################
    # Forward methods
    ############################
    def add_question(self, other: Question) -> "Survey":
        """Add a question to this question by turning them into a survey with two questions."""
        from edsl.surveys.Survey import Survey

        s = Survey([self, other])
        return s

    def run(self, *args, **kwargs):
        """Turn a single question into a survey and run it."""
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s.run(*args, **kwargs)

    def by(self, *args):
        """Turn a single question into a survey and run it."""
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s.by(*args)

    def human_readable(self):
        """Print the question in a human readable format."""
        lines = []
        lines.append(f"Question Type: {self.question_type}")
        lines.append(f"Question: {self.question_text}")
        if hasattr(self, "question_options"):
            lines.append("Please name the option you choose from the following.:")
            for index, option in enumerate(self.question_options):
                lines.append(f"{option}")
        return "\n".join(lines)

    def rich_print(self):
        """Print the question in a rich format."""
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Question Name", style="dim")
        table.add_column("Question Type")
        table.add_column("Question Text")
        table.add_column("Options")

        question = self
        if hasattr(question, "question_options"):
            options = ", ".join([str(o) for o in question.question_options])
        else:
            options = "None"
        table.add_row(
            question.question_name,
            question.question_type,
            question.question_text,
            options,
        )
        return table


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.inference_services.InferenceServicesCollection import (
    InferenceServicesCollection,
)

from edsl.inference_services.OpenAIService import OpenAIService
from edsl.inference_services.AnthropicService import AnthropicService
from edsl.inference_services.DeepInfraService import DeepInfraService
from edsl.inference_services.GoogleService import GoogleService

default = InferenceServicesCollection(
    [OpenAIService, AnthropicService, DeepInfraService, GoogleService]
)



from typing import Union, List, Any
import asyncio
import time
from collections import UserDict
from matplotlib import pyplot as plt


class TokenBucket:
    """This is a token bucket used to respect rate limits to services."""

    def __init__(
        self,
        *,
        bucket_name: str,
        bucket_type: str,
        capacity: Union[int, float],
        refill_rate: Union[int, float],
    ):
        self.bucket_name = bucket_name
        self.bucket_type = bucket_type
        self.capacity = capacity  # Maximum number of tokens
        self.tokens = capacity  # Current number of available tokens
        self.refill_rate = refill_rate  # Rate at which tokens are refilled
        self.last_refill = time.monotonic()  # Last refill time

        self.log: List[Any] = []

    def __add__(self, other) -> "TokenBucket":
        """Combine two token buckets.

        The resulting bucket has the minimum capacity and refill rate of the two buckets.
        This is useful, for example, if we have two calls to the same model on the same service but have different temperatures.
        """
        return TokenBucket(
            bucket_name=self.bucket_name,
            bucket_type=self.bucket_type,
            capacity=min(self.capacity, other.capacity),
            refill_rate=min(self.refill_rate, other.refill_rate),
        )

    def __repr__(self):
        return f"TokenBucket(bucket_name={self.bucket_name}, bucket_type='{self.bucket_type}', capacity={self.capacity}, refill_rate={self.refill_rate})"

    def add_tokens(self, tokens: Union[int, float]) -> None:
        """Add tokens to the bucket, up to the maximum capacity."""
        self.tokens = min(self.capacity, self.tokens + tokens)
        self.log.append((time.monotonic(), self.tokens))

    def refill(self) -> None:
        """Refill the bucket with new tokens based on elapsed time."""
        now = time.monotonic()
        elapsed = now - self.last_refill
        refill_amount = elapsed * self.refill_rate
        self.tokens = min(self.capacity, self.tokens + refill_amount)
        self.last_refill = now

        self.log.append((now, self.tokens))

    def wait_time(self, requested_tokens: Union[float, int]) -> float:
        """Calculate the time to wait for the requested number of tokens."""
        now = time.monotonic()
        elapsed = now - self.last_refill
        refill_amount = elapsed * self.refill_rate
        available_tokens = min(self.capacity, self.tokens + refill_amount)
        return max(0, requested_tokens - available_tokens) / self.refill_rate

    async def get_tokens(self, amount: Union[int, float] = 1, warn=True) -> None:
        """Wait for the specified number of tokens to become available.
        Note that this method is a coroutine.
        """
        if amount > self.capacity:
            msg = f"Requested amount exceeds bucket capacity. Bucket capacity: {self.capacity}, requested amount: {amount}. As the bucket never overflows, the requested amount will never be available."
            raise ValueError(msg)
        while self.tokens < amount:
            self.refill()
            await asyncio.sleep(0.1)  # Sleep briefly to prevent busy waiting
        self.tokens -= amount

        now = time.monotonic()
        self.log.append((now, self.tokens))

    def get_log(self) -> list[tuple]:
        return self.log

    def visualize(self):
        """Visualize the token bucket over time."""
        times, tokens = zip(*self.get_log())
        start_time = times[0]
        times = [t - start_time for t in times]  # Normalize time to start from 0

        plt.figure(figsize=(10, 6))
        plt.plot(times, tokens, label="Tokens Available")
        plt.xlabel("Time (seconds)", fontsize=12)
        plt.ylabel("Number of Tokens", fontsize=12)
        details = f"{self.bucket_name} ({self.bucket_type}) Bucket Usage Over Time\nCapacity: {self.capacity:.1f}, Refill Rate: {self.refill_rate:.1f}/second"
        plt.title(details, fontsize=14)

        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.show()



from textwrap import dedent


class LanguageModelExceptions(Exception):
    pass


class LanguageModelNotFound(LanguageModelExceptions):
    def __init__(self, model_name):
        msg = dedent(
            f"""\
            Model {model_name} not found.
            To create an instance, you can do: 
            >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
            
            To get the default model, you can leave out the model name. 
            To see the available models, you can do:
            >>> Model.available()

            See https://docs.expectedparrot.com/en/latest/language_models.html#available-models for more details.
            """
        )
        super().__init__(msg)


class LanguageModelResponseNotJSONError(LanguageModelExceptions):
    pass


class LanguageModelMissingAttributeError(LanguageModelExceptions):
    pass


class LanguageModelAttributeTypeError(LanguageModelExceptions):
    pass


class LanguageModelDoNotAddError(LanguageModelExceptions):
    pass



import aiohttp
import json
import os
import requests
from inspect import isclass
from requests.exceptions import ConnectionError
from typing import Any, Optional, Type, Union, Literal
from uuid import UUID
import edsl
from edsl import CONFIG
from edsl.agents import Agent, AgentList
from edsl.config import CONFIG
from edsl.data.Cache import Cache
from edsl.questions.QuestionBase import QuestionBase
from edsl.results import Results
from edsl.surveys import Survey
from edsl.data.CacheEntry import CacheEntry


class Coop:
    """
    Client for the Expected Parrot API.
    """

    def __init__(self, api_key: str = None, url: str = None) -> None:
        """
        Initialize the client.
        """
        self.api_key = api_key or os.getenv("EXPECTED_PARROT_API_KEY")
        self.url = url or CONFIG.EXPECTED_PARROT_URL

    ################
    # BASIC METHODS
    ################
    @property
    def headers(self) -> dict:
        """
        Return the headers for the request.
        """
        headers = {}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        return headers

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Send a request to the server and return the response.
        """
        url = f"{self.url}/{uri}"
        try:
            if method.upper() in ["GET", "DELETE"]:
                response = requests.request(
                    method, url, params=params, headers=self.headers
                )
            else:
                response = requests.request(
                    method, url, json=payload, headers=self.headers
                )
        except ConnectionError:
            raise ConnectionError("Could not connect to the server.")

        return response

    def _resolve_server_response(self, response: requests.Response) -> None:
        """
        Check the response from the server and raise appropriate errors.
        """
        if response.status_code >= 400:
            message = response.json().get("detail")
            if "Authorization" in message:
                message = "Please provide an Expected Parrot API key."
            raise Exception(message)

    ################
    # HELPER METHODS
    ################
    async def remote_async_execute_model_call(
        self, model_dict: dict, user_prompt: str, system_prompt: str
    ) -> dict:
        url = self.url + "/inference/"
        # print("Now using url: ", url)
        data = {
            "model_dict": model_dict,
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }
        # Use aiohttp to send a POST request asynchronously
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=data) as response:
                response_data = await response.json()
        return response_data

    def web(
        self,
        survey: dict,
        platform: Literal[
            "google_forms", "lime_survey", "survey_monkey"
        ] = "lime_survey",
        email=None,
    ):
        url = f"{self.url}/api/v0/export_to_{platform}"
        if email:
            data = {"json_string": json.dumps({"survey": survey, "email": email})}
        else:
            data = {"json_string": json.dumps({"survey": survey, "email": email})}

        response_json = requests.post(url, data=json.dumps(data))

        return response_json

    def _json_handle_none(self, value: Any) -> Any:
        """
        Helper function to handle None values in JSON serialization.
        - Returns "null" if value is None. If not, doesn't return anything else.
        """
        if value is None:
            return "null"

    def _get_edsl_version(self) -> str:
        """
        Return the version of the EDSL.
        """
        try:
            version = edsl.__version__
        except:
            version = None
        return version

    def _resolve_edsl_object(
        self,
        object: Union[
            Type[QuestionBase], QuestionBase, Survey, Agent, AgentList, Results
        ],
    ) -> tuple[str, str]:
        """
        Resolves an EDSL object or class to an (API_URI: str, object_type: str) tuple.
        """
        object_type = type(object) if not isclass(object) else object

        if issubclass(object_type, QuestionBase):
            return ("questions", "question")
        elif issubclass(object_type, Survey):
            return ("surveys", "survey")
        elif issubclass(object_type, Agent) or issubclass(object_type, AgentList):
            return ("agents", "agent")
        elif issubclass(object_type, Results):
            return ("results", "results")
        elif issubclass(object_type, Cache):
            return ("caches", "cache")
        else:
            raise ValueError("Incorrect or not supported object type")

    def _resolve_object_type(self, object_type: str) -> tuple[str, Type]:
        """
        Resolve an object_type string to an (API_URI: str, object_type: Type) tuple.
        """
        TYPE_MAP = {
            "question": ("questions", QuestionBase),
            "survey": ("surveys", Survey),
            "agent": ("agents", Agent),
            "results": ("results", Results),
            "cache": ("caches", Cache),
        }

        if object_type is None:
            raise ValueError("Please provide an `object_type`.")
        elif object_type not in TYPE_MAP:
            raise ValueError(
                f"Object type {object_type} not recognized. "
                f"Valid object types are: {', '.join(TYPE_MAP.keys())}"
            )

        return TYPE_MAP[object_type]

    def _resolve_uri(self, uri: str) -> tuple[str, Type]:
        """
        Resolve an API URI to an (API_URI: str, object_type: Type) tuple.
        """
        URI_MAP = {
            "questions": ("question", QuestionBase),
            "surveys": ("survey", Survey),
            "agents": ("agent", Agent),
            "results": ("results", Results),
            "caches": ("cache", Cache),
        }

        if uri is None:
            raise ValueError("Please provide a URI.")
        elif uri not in URI_MAP:
            raise ValueError(f"URI {uri} not recognized")

        return URI_MAP[uri]

    ################
    # OBJECT METHODS
    ################

    # ----------
    # A. CREATE
    # ----------
    def _create(
        self,
        object: Union[Type[QuestionBase], Survey, Agent, AgentList, Results],
        visibility: Literal["public", "unlisted", "private"] = "unlisted",
    ) -> dict:
        """
        Create an EDSL object in the Coop server.

        :param object: the EDSL object to be sent.
        :param visibility: the object's visibility (defaults to "unlisted").
        """

        uri, _ = self._resolve_edsl_object(object)
        response = self._send_server_request(
            uri=f"api/v0/{uri}",
            method="POST",
            payload={
                "json_string": json.dumps(
                    object.to_dict(), default=self._json_handle_none
                ),
                "visibility": visibility,
                "version": self._get_edsl_version(),
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        url = f"{self.url}/explore/{uri}/{response_json['uuid']}"
        response_json["url"] = url
        return response_json

    def create(
        self,
        object: Union[Type[QuestionBase], Survey, Agent, AgentList, Results],
        visibility: Literal["public", "unlisted", "private"] = "unlisted",
        verbose: bool = False,
    ) -> dict:
        """
        Create an EDSL object in the Coop server.

        :param object: the EDSL object to be sent.
        :param visibility: the object's visibility (defaults to "unlisted").
        """
        response = self._create(object, visibility)
        if verbose:
            print(f"Object pushed to Coop - available at {response['url']}")
        return response

    # ----------
    # B. GET
    # ----------
    def _get(self, object_type_uri: str, uuid: Union[str, UUID]) -> dict:
        """
        Retrieve an EDSL object from the Coop server.
        """
        response = self._send_server_request(
            uri=f"api/v0/{object_type_uri}/{uuid}", method="GET"
        )
        self._resolve_server_response(response)
        return json.loads(response.json().get("json_string"))

    def get(
        self, object_type: str = None, uuid: Union[str, UUID] = None, url: str = None
    ) -> Union[Type[QuestionBase], Survey, Agent, AgentList, Results]:
        """
        Retrieve an EDSL object by its object type and UUID, or by its url.

        :param object_type: the type of object to retrieve.
        :param uuid: the uuid of the object either in str or UUID format.
        :param url: the url of the object.
        """
        if url:
            uri = url.split("/")[-2]
            object_type, cls = self._resolve_uri(uri)
            uuid = url.split("/")[-1]
        elif object_type and uuid:
            uri, cls = self._resolve_object_type(object_type)
        else:
            raise ValueError(
                "Please provide either an object type and a UUID, or a url."
            )

        json_dict = self._get(object_type_uri=uri, uuid=uuid)
        if object_type == "agent" and "agent_list" in json_dict:
            return AgentList.from_dict(json_dict)
        else:
            return cls.from_dict(json_dict)

    def _get_base(
        self,
        cls: Union[Type[QuestionBase], Survey, Agent, AgentList, Results],
        uuid: Union[str, UUID],
    ):
        """
        Used by the Base class to offer a get functionality.
        """
        _, object_type = self._resolve_edsl_object(cls)
        return self.get(object_type, uuid)

    # ----------
    # C. GET ALL
    # ----------
    @property
    def questions(self) -> list[dict[str, Union[int, QuestionBase]]]:
        """Retrieve all Questions."""
        response = self._send_server_request(uri="api/v0/questions", method="GET")
        self._resolve_server_response(response)
        questions = [
            {
                "question": QuestionBase.from_dict(json.loads(q["json_string"])),
                "uuid": q["uuid"],
                "version": q["version"],
                "visibility": q.get("visibility"),
                "url": f"{self.url}/explore/questions/{q['uuid']}",
            }
            for q in response.json()
        ]
        return questions

    @property
    def surveys(self) -> list[dict[str, Union[int, Survey]]]:
        """Retrieve all Surveys."""
        response = self._send_server_request(uri="api/v0/surveys", method="GET")
        self._resolve_server_response(response)
        surveys = [
            {
                "survey": Survey.from_dict(json.loads(s["json_string"])),
                "uuid": s["uuid"],
                "version": s["version"],
                "visibility": s.get("visibility"),
                "url": f"{self.url}/explore/surveys/{s['uuid']}",
            }
            for s in response.json()
        ]
        return surveys

    @property
    def agents(self) -> list[dict[str, Union[int, Agent, AgentList]]]:
        """Retrieve all Agents and AgentLists."""
        response = self._send_server_request(uri="api/v0/agents", method="GET")
        self._resolve_server_response(response)
        agents = []
        for a in response.json():
            agent_dict = json.loads(a.get("json_string"))
            if "agent_list" in agent_dict:
                agent = AgentList.from_dict(agent_dict)
            else:
                agent = Agent.from_dict(agent_dict)
            agents.append(
                {
                    "uuid": a.get("uuid"),
                    "agent": agent,
                    "version": a.get("version"),
                    "visibility": a.get("visibility"),
                    "url": f"{self.url}/explore/agents/{a['uuid']}",
                }
            )
        return agents

    @property
    def results(self) -> list[dict[str, Union[int, Results]]]:
        """Retrieve all Results."""
        response = self._send_server_request(uri="api/v0/results", method="GET")
        self._resolve_server_response(response)
        results = [
            {
                "uuid": r.get("uuid"),
                "results": Results.from_dict(json.loads(r.get("json_string"))),
                "version": r.get("version"),
                "visibility": r.get("visibility"),
                "url": f"{self.url}/explore/results/{r['uuid']}",
            }
            for r in response.json()
        ]
        return results

    @property
    def caches(self) -> list[dict[str, Union[int, Cache]]]:
        """Retrieve all Caches."""
        response = self._send_server_request(uri="api/v0/caches", method="GET")
        self._resolve_server_response(response)
        caches = [
            {
                "uuid": c.get("uuid"),
                "cache": Cache.from_dict(json.loads(c.get("json_string"))),
                "version": c.get("version"),
                "visibility": c.get("visibility"),
                "url": f"{self.url}/explore/caches/{c['uuid']}",
            }
            for c in response.json()
        ]
        return caches

    # ----------
    # D. DELETE
    # ----------
    def delete(self, object_type: str, uuid: Union[str, UUID]) -> dict:
        """
        Delete an EDSL object from the Coop server.

        :param object_type: the type of object to delete.
        :param uuid: the uuid of the object either in str or UUID format.
        """

        uri, _ = self._resolve_object_type(object_type)
        response = self._send_server_request(
            uri=f"api/v0/{uri}/{uuid}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # CacheEntry Methods
    ################
    def create_cache_entry(
        self, cache_entry: CacheEntry, visibility: str = "unlisted"
    ) -> dict:
        """
        Create a CacheEntry object.
        """
        response = self._send_server_request(
            uri="api/v0/cache-entries",
            method="POST",
            payload={
                "visibility": visibility,
                "version": self._get_edsl_version(),
                "json_string": json.dumps(
                    {"key": cache_entry.key, "value": json.dumps(cache_entry.to_dict())}
                ),
            },
        )
        self._resolve_server_response(response)
        return response.json()

    def create_cache_entries(
        self, cache_entries: dict[str, CacheEntry], visibility: str = "unlisted"
    ) -> None:
        """
        Send a dictionary of CacheEntry objects to the server.
        """
        response = self._send_server_request(
            uri="api/v0/cache-entries/many",
            method="POST",
            payload={
                "visibility": visibility,
                "version": self._get_edsl_version(),
                "json_string": json.dumps(
                    {k: json.dumps(v.to_dict()) for k, v in cache_entries.items()}
                ),
            },
        )
        # print(response.json())
        self._resolve_server_response(response)
        return response.json()

    def get_cache_entries(
        self, exclude_keys: Optional[list[str]] = None
    ) -> list[CacheEntry]:
        """
        Return CacheEntry objects from the server.

        :param exclude_keys: exclude CacheEntry objects with these keys.
        """
        if exclude_keys is None:
            exclude_keys = []
        response = self._send_server_request(
            uri="api/v0/cache-entries/get-many",
            method="POST",
            payload={"keys": exclude_keys},
        )
        self._resolve_server_response(response)
        return [
            CacheEntry.from_dict(json.loads(v.get("json_string")))
            for v in response.json()
        ]

    ################
    # Error Message Methods
    ################
    def send_error_message(self, error_data: str) -> dict:
        """
        Send an error message to the server.
        """
        response = self._send_server_request(
            uri="api/v0/errors",
            method="POST",
            payload={"json_string": json.dumps(error_data)},
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # DUNDER METHODS
    ################
    def __repr__(self):
        """Return a string representation of the client."""
        return f"Client(api_key='{self.api_key}', url='{self.url}')"


if __name__ == "__main__":
    from edsl.coop import Coop
    import uuid

    API_KEY = "b"
    coop = Coop(api_key=API_KEY)

    # basics
    coop

    ##############
    # A. Objects
    ##############

    # ------------
    # A.1 Questions
    # ------------
    from edsl.questions import QuestionMultipleChoice
    from edsl.questions import QuestionCheckBox
    from edsl.questions import QuestionFreeText

    # check questions on server (should be an empty list)
    coop.questions
    for item in coop.questions:
        coop.delete(object_type="question", uuid=item.get("uuid"))
    # try to get a question that does not exist - should get an error
    coop.get(object_type="question", uuid=uuid.uuid4())
    coop.get(object_type="question", uuid=str(uuid.uuid4()))
    # now post some questions
    response = coop.create(QuestionMultipleChoice.example())
    coop.create(QuestionCheckBox.example(), visibility="private")
    coop.create(QuestionFreeText.example(), visibility="public")
    # check all questions - there must be three
    coop.questions
    # or get a question by its uuid
    coop.get(object_type="question", uuid=response.get("uuid"))
    # or by its url
    coop.get(url=response.get("url"))
    # delete the question
    coop.delete(object_type="question", uuid=response.get("uuid"))
    # check all questions - there must be two left
    coop.questions

    # ------------
    # A.2 Surveys
    # ------------
    from edsl.surveys import Survey

    # check surveys on server (should be an empty list)
    coop.surveys
    for survey in coop.surveys:
        coop.delete(object_type="survey", uuid=survey.get("uuid"))
    # try to get a survey that does not exist - should get an error
    coop.get(object_type="survey", uuid=uuid.uuid4())
    coop.get(object_type="survey", uuid=str(uuid.uuid4()))
    # now post some surveys
    response = coop.create(Survey.example())
    coop.create(Survey.example(), visibility="private")
    coop.create(Survey.example(), visibility="public")
    s = Survey().example()
    for i in range(10):
        q = QuestionFreeText.example()
        q.question_name = f"question_{i}"
        s.add_question(q)
    coop.create(s, visibility="public")
    # check all surveys - there must be three
    coop.surveys
    # or get survey by uuid
    coop.get(object_type="survey", uuid=response.get("uuid"))
    # or by its url
    coop.get(url=response.get("url"))
    # delete the survey
    coop.delete(object_type="survey", uuid=response.get("uuid"))
    # check all surveys - there must be two left
    coop.surveys

    # ------------
    # A.3 Agents
    # ------------
    from edsl.agents import Agent, AgentList

    # check agents on server (should be an empty list)
    coop.agents
    for agent in coop.agents:
        coop.delete(object_type="agent", uuid=agent.get("uuid"))
    # try to get an agent that does not exist - should get an error
    coop.get(object_type="agent", uuid=uuid.uuid4())
    coop.get(object_type="agent", uuid=str(uuid.uuid4()))
    # now post some agents
    response = coop.create(Agent.example())
    coop.create(Agent.example(), visibility="private")
    coop.create(Agent.example(), visibility="public")
    coop.create(
        Agent(traits={"hair_type": "curly", "skil_color": "white"}), visibility="public"
    )
    coop.create(AgentList.example())
    coop.create(AgentList.example(), visibility="private")
    coop.create(AgentList.example(), visibility="public")
    # check all agents - there must be a few
    coop.agents
    # or get agent by uuid
    coop.get(object_type="agent", uuid=response.get("uuid"))
    # or by its url
    coop.get(url=response.get("url"))
    # delete the agent
    coop.delete(object_type="agent", uuid=response.get("uuid"))
    # check all agents
    coop.agents

    # ------------
    # A.4 Results
    # ------------
    from edsl.results import Results

    # check results on server (should be an empty list)
    coop.results
    for results in coop.results:
        coop.delete(object_type="results", uuid=results.get("uuid"))
    # try to get a results that does not exist - should get an error
    coop.get(object_type="results", uuid=uuid.uuid4())
    coop.get(object_type="results", uuid=str(uuid.uuid4()))
    # now post some Results
    response = coop.create(Results.example())
    coop.create(Results.example(), visibility="private")
    coop.create(Results.example(), visibility="public")
    # check all results - there must be a few
    coop.results
    # or get results by uuid
    coop.get(object_type="results", uuid=response.get("uuid"))
    # or by its url
    coop.get(url=response.get("url"))
    # delete the results
    coop.delete(object_type="results", uuid=response.get("uuid"))
    # check all results
    coop.results

    # ------------
    # A.5 Caches
    # ------------
    from edsl.data import Cache

    # check caches on server (should be an empty list)
    coop.caches
    for cache in coop.caches:
        coop.delete(object_type="cache", uuid=cache.get("uuid"))
    # try to get a cache that does not exist - should get an error
    coop.get(object_type="cache", uuid=uuid.uuid4())
    coop.get(object_type="cache", uuid=str(uuid.uuid4()))
    # now post some Caches
    response = coop.create(Cache.example())
    coop.create(Cache.example(), visibility="private")
    coop.create(Cache.example(), visibility="public")
    # check all caches - there must be a few
    coop.caches
    # or get cache by uuid
    coop.get(object_type="cache", uuid=response.get("uuid"))
    # or by its url
    coop.get(url=response.get("url"))
    # delete the cache
    coop.delete(object_type="cache", uuid=response.get("uuid"))
    # check all caches
    coop.caches

    ##############
    # B. CacheEntries
    ##############
    from edsl.data.CacheEntry import CacheEntry

    # should be empty in the beginning
    coop.get_cache_entries()
    # now create one cache entry
    cache_entry = CacheEntry.example()
    coop.create_cache_entry(cache_entry)
    # see that if you try to create it again, you'll get the same uuid
    coop.create_cache_entry(cache_entry)
    # now get all your cache entries
    coop.get_cache_entries()
    coop.get_cache_entries(exclude_keys=[])
    coop.get_cache_entries(exclude_keys=["a"])
    # this will be empty
    coop.get_cache_entries(exclude_keys=[cache_entry.key])
    # now send many cache entries
    cache_entries = {}
    for i in range(10):
        cache_entry = CacheEntry.example(randomize=True)
        cache_entries[cache_entry.key] = cache_entry
    coop.create_cache_entries(cache_entries)

    ##############
    # E. ERROR MESSAGE
    ##############
    coop = Coop()
    coop.api_key = "a"
    coop.send_error_message({"something": "This is an error message"})
    coop.api_key = None
    coop.send_error_message({"something": "This is an error message"})



from __future__ import annotations
from typing import List, DefaultDict
import asyncio

from rich.text import Text
from rich.box import SIMPLE
from rich.table import Table

from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.runners.JobsRunnerStatusData import JobsRunnerStatusData

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]


class JobsRunnerStatusPresentation:
    @staticmethod
    def display_status_table(status_summary):
        table = Table(
            title="Job Status",
            show_header=True,
            header_style="bold magenta",
            box=SIMPLE,
        )
        table.max_width = 100
        table.add_column("Statistic", style="dim", no_wrap=True, width=50)
        table.add_column("Value", width=10)

        for key, value in status_summary.items():
            if key != "model_queues":
                table.add_row(key, value)

        spacing = " "
        if "model_queues" in status_summary:
            table.add_row(Text("Model Queues", style="bold red"), "")
            for model_info in status_summary["model_queues"]:
                model_name = model_info["model_name"]
                tpm = "TPM (k)=" + str(model_info["TPM_limit_k"])
                rpm = "RPM (k)=" + str(model_info["RPM_limit_k"])
                pretty_model_name = model_name + ";" + tpm + ";" + rpm
                table.add_row(Text(pretty_model_name, style="blue"), "")
                table.add_row(
                    "Number question tasks waiting for capacity",
                    str(model_info["num_tasks_waiting"]),
                )
                # Token usage and cost info
                for cache_info in model_info["token_usage_info"]:
                    cache_status = cache_info["cache_status"]
                    table.add_row(
                        Text(spacing + cache_status.replace("_", " "), style="bold"), ""
                    )
                    for detail in cache_info["details"]:
                        token_type = detail["type"]
                        tokens = detail["tokens"]
                        #                 cost = detail["cost"]
                        table.add_row(spacing + f"{token_type}", f"{tokens:,}")
                    table.add_row(spacing + "cost", cache_info["cost"])

        return table


class JobsRunnerStatusMixin(JobsRunnerStatusData, JobsRunnerStatusPresentation):
    def status_data(self, completed_tasks: List[asyncio.Task], elapsed_time: float):
        # return self.generate_status_summary(
        #     completed_tasks=completed_tasks,
        #     elapsed_time=elapsed_time,
        #     interviews=self.total_interviews).rawplt.figure(figsize=(10, 6))

        # return self.full_status(self.total_interviews)
        return None

    def status_table(self, completed_tasks: List[asyncio.Task], elapsed_time: float):
        summary_data = self.generate_status_summary(
            completed_tasks=completed_tasks,
            elapsed_time=elapsed_time,
            interviews=self.total_interviews,
        )
        return self.display_status_table(summary_data)



from __future__ import annotations
import textwrap
from random import uniform
from typing import Any, Optional, Union
from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import NumericalOrNoneDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionNumerical(QuestionBase):
    """This question prompts the agent to answer with a numerical value."""

    question_type = "numerical"
    min_value: Optional[float] = NumericalOrNoneDescriptor()
    max_value: Optional[float] = NumericalOrNoneDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionNumerical.default_instructions`.
        :param min_value: The minimum value of the answer.
        :param max_value: The maximum value of the answer.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.min_value = min_value
        self.max_value = max_value

    ################
    # Answer methods
    ################
    def _validate_answer(
        self, answer: dict[str, Any]
    ) -> dict[str, Union[str, float, int]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value_numeric(answer, "answer")
        self._validate_answer_numerical(answer)
        return answer

    def _translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """There is no answer code."""
        return answer

    def _simulate_answer(self, human_readable: bool = True):
        """Simulate a valid answer for debugging purposes."""
        return {
            "answer": uniform(self.min_value, self.max_value),
            "comment": random_string(),
        }

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionNumerical:
        """Return an example question."""
        return cls(
            question_name="age",
            question_text="How old are you in years?",
            min_value=0,
            max_value=86.7,
        )


def main():
    """Show example usage."""
    from edsl.questions.QuestionNumerical import QuestionNumerical

    q = QuestionNumerical.example()
    q.question_text
    q.min_value
    q.max_value
    # validate an answer
    q._validate_answer({"answer": 1, "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer(1)
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""This module contains the Answers class, which is a helper class to hold the answers to a survey."""
from collections import UserDict
from rich.table import Table


class Answers(UserDict):
    """Helper class to hold the answers to a survey."""

    def add_answer(self, response, question) -> None:
        """Add a response to the answers dictionary."""
        answer = response.get("answer")
        comment = response.pop("comment", None)
        # record the answer
        self[question.question_name] = answer
        if comment:
            self[question.question_name + "_comment"] = comment

    def replace_missing_answers_with_none(self, survey) -> None:
        """Replace missing answers with None. Answers can be missing if the agent skips a question."""
        for question_name in survey.question_names:
            if question_name not in self:
                self[question_name] = None

    def to_dict(self):
        """Return a dictionary of the answers."""
        return self.data

    @classmethod
    def from_dict(cls, d):
        """Return an Answers object from a dictionary."""
        return cls(d)

    def rich_print(self):
        """Display an object as a table."""
        table = Table(title="Answers")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))

        return table



from edsl import CONFIG

from tenacity import (
    retry,
    wait_exponential,
    stop_after_attempt,
    retry_if_exception_type,
    before_sleep,
)

EDSL_BACKOFF_START_SEC = float(CONFIG.get("EDSL_BACKOFF_START_SEC"))
EDSL_MAX_BACKOFF_SEC = float(CONFIG.get("EDSL_MAX_BACKOFF_SEC"))
EDSL_MAX_ATTEMPTS = int(CONFIG.get("EDSL_MAX_ATTEMPTS"))


def print_retry(retry_state, print_to_terminal=False):
    "Prints details on tenacity retries."
    attempt_number = retry_state.attempt_number
    exception = retry_state.outcome.exception()
    wait_time = retry_state.next_action.sleep
    # breakpoint()
    if print_to_terminal:
        print(
            f"Attempt {attempt_number} failed with exception: {repr(exception)}; "
            f"now waiting {wait_time:.2f} seconds before retrying."
            f" Parameters: start={EDSL_BACKOFF_START_SEC}, max={EDSL_MAX_BACKOFF_SEC}, max_attempts={EDSL_MAX_ATTEMPTS}."
        )


retry_strategy = retry(
    wait=wait_exponential(
        multiplier=EDSL_BACKOFF_START_SEC, max=EDSL_MAX_BACKOFF_SEC
    ),  # Exponential back-off starting at 1s, doubling, maxing out at 60s
    stop=stop_after_attempt(EDSL_MAX_ATTEMPTS),  # Stop after 5 attempts
    # retry=retry_if_exception_type(Exception),  # Customize this as per your specific retry-able exception
    before_sleep=print_retry,  # Use custom print function for retries
)



"""A list of Agent objects.

Example usage:

.. code-block:: python

    al = AgentList([Agent.example(), Agent.example()])
    len(al)
    2
    
"""
from __future__ import annotations
from collections import UserList
from typing import Optional, Union

from rich.table import Table

from edsl.Base import Base
from edsl.agents import Agent
from edsl.agents.AgentListExportMixin import AgentListExportMixin


class AgentList(UserList, Base, AgentListExportMixin):
    """A list of Agents."""

    def __init__(self, data: Optional[list[Agent]] = None):
        """Initialize a new AgentList.

        :param data: A list of Agents.
        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def to(self, question_or_survey: Union["Question", "Survey"]) -> "Jobs":
        """Return a Job with a question or survey taken by the agent."""
        return question_or_survey.by(*self)

    def to_dict(self):
        """Return dictionary of AgentList to serialization."""
        return {"agent_list": [agent.to_dict() for agent in self.data]}

    def _repr_html_(self):
        """Return an HTML representation of the AgentList."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict()["agent_list"])

    @classmethod
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserialize the dictionary back to an AgentList object.

        :param: data: A dictionary representing an AgentList.
        """
        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls) -> "AgentList":
        """Return an example AgentList."""
        return cls([Agent.example(), Agent.example()])

    def code(self) -> list[str]:
        """Return code to construct an AgentList."""
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        return lines

    def rich_print(self) -> Table:
        """Display an object as a rich table."""
        table = Table(title="AgentList")
        table.add_column("Agents", style="bold")
        for agent in self.data:
            table.add_row(agent.rich_print())
        return table


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Module to store system information."""
from dataclasses import dataclass
import getpass
import platform
import pkg_resources


@dataclass
class SystemInfo:
    """Dataclass to store system information."""

    username: str
    system_info: str
    release_info: str
    package_name: str
    package_version: str

    def __init__(self, package_name: str):
        """Initialize the dataclass with system."""
        self.username = getpass.getuser()
        self.system_info = platform.system()
        self.release_info = platform.release()
        self.package_name = package_name
        try:
            self.package_version = pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            self.package_version = "Not installed"



from edsl.jobs.tasks.task_status_enum import TaskStatus
from matplotlib import pyplot as plt
from typing import List


class TaskHistory:
    def __init__(self, interviews: List["Interview"], include_traceback=False):
        self.total_interviews = interviews
        self.include_traceback = include_traceback

        self.exceptions = [
            i.exceptions
            for index, i in enumerate(self.total_interviews)
            if i.exceptions != {}
        ]
        self.indices = [
            index for index, i in enumerate(self.total_interviews) if i.exceptions != {}
        ]

    def to_dict(self):
        """Return the TaskHistory as a dictionary."""
        return {
            "exceptions": [
                e.to_dict(include_traceback=self.include_traceback)
                for e in self.exceptions
            ],
            "indices": self.indices,
        }

    @property
    def has_exceptions(self) -> bool:
        """Return True if there are any exceptions."""
        return len(self.exceptions) > 0

    def _repr_html_(self):
        """Return an HTML representation of the TaskHistory."""
        from edsl.utilities.utilities import data_to_html

        newdata = self.to_dict()["exceptions"]
        return data_to_html(newdata, replace_new_lines=True)

    def show_exceptions(self):
        """Print the exceptions."""
        for index in self.indices:
            self.total_interviews[index].exceptions.print()

    def get_updates(self):
        """Return a list of all the updates."""
        updates = []
        for interview in self.total_interviews:
            for question_name, logs in interview.task_status_logs.items():
                updates.append(logs)
        return updates

    def print(self):
        from rich import print

        print(self.get_updates())

    def plot_completion_times(self):
        """Plot the completion times for each task."""
        updates = self.get_updates()

        elapsed = [update.max_time - update.min_time for update in updates]
        for i, update in enumerate(updates):
            if update[-1]["value"] != TaskStatus.SUCCESS:
                elapsed[i] = 0
        x = range(len(elapsed))
        y = elapsed

        plt.bar(x, y)
        plt.title("Per-interview completion times")
        plt.xlabel("Task")
        plt.ylabel("Time (seconds)")
        plt.show()

    def plotting_data(self, num_periods=100):
        updates = self.get_updates()

        min_t = min([update.min_time for update in updates])
        max_t = max([update.max_time for update in updates])
        delta_t = (max_t - min_t) / (num_periods * 1.0)
        time_periods = [min_t + delta_t * i for i in range(num_periods)]

        def counts(t):
            d = {}
            for update in updates:
                status = update.status_at_time(t)
                if status in d:
                    d[status] += 1
                else:
                    d[status] = 1
            return d

        status_counts = [counts(t) for t in time_periods]

        new_counts = []
        for status_count in status_counts:
            d = {task_status: 0 for task_status in TaskStatus}
            d.update(status_count)
            new_counts.append(d)

        return new_counts

    def plot(self, num_periods=100):
        """Plot the number of tasks in each state over time."""
        new_counts = self.plotting_data(num_periods)
        max_count = max([max(entry.values()) for entry in new_counts])

        rows = int(len(TaskStatus) ** 0.5) + 1
        cols = (len(TaskStatus) + rows - 1) // rows  # Ensure all plots fit
        from matplotlib import pyplot as plt

        plt.figure(figsize=(15, 10))  # Adjust the figure size as needed
        for i, status in enumerate(TaskStatus, start=1):
            plt.subplot(rows, cols, i)
            x = range(len(new_counts))
            y = [
                item.get(status, 0) for item in new_counts
            ]  # Use .get() to handle missing keys safely
            plt.plot(x, y, marker="o", linestyle="-")
            plt.title(status.name)
            plt.xlabel("Time Periods")
            plt.ylabel("Count")
            plt.grid(True)
            plt.ylim(0, max_count)

        plt.tight_layout()
        plt.show()



"""This module contains the Memory class, which is a list of prior questions."""
from collections import UserList


class Memory(UserList):
    """Class for holding the questions (stored as names) that we want the the agent to have available when answering a question."""

    def __init__(self, prior_questions: list[str] = None):
        """Initialize the Memory object."""
        super().__init__(prior_questions or [])

    def add_prior_question(self, prior_question):
        """Add a prior question to the memory."""
        if prior_question not in self:
            self.append(prior_question)
        else:
            raise ValueError(f"{prior_question} is already in the memory.")

    def __repr__(self):
        """Return a string representation of the Memory object."""
        return f"Memory(prior_questions={self.data})"

    def to_dict(self):
        """Create a dictionary representation of the Memory object."""
        return {"prior_questions": self.data}

    @classmethod
    def from_dict(cls, data):
        """Create a Memory object from a dictionary."""
        return cls(**data)
